from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db import models, router
from django.db.models.fields.related_descriptors import ForwardManyToOneDescriptor

from isapilib.api.models import UserAPI  # noqa keep this
from isapilib.logging import logger


class DynamicManager(models.Manager):
    def get_queryset(self):
        queryset = super().get_queryset()
        meta = queryset.model._meta
        db_table_options = meta.db_table

        if not isinstance(db_table_options, dict):
            return queryset

        name = queryset.model.__name__
        using = router.db_for_write(None)
        version = settings.DATABASES[using]['INTELISIS_VERSION']

        try:
            db_table_options = {int(k): v for k, v in db_table_options.items()}
        except ValueError:
            raise ImproperlyConfigured(f'{name} does not have a valid db_table definition, expected {{int:str}}')

        db_table_default = getattr(meta, 'db_table_default', 6000)

        logger.debug('Dynamic table resolution for model %s using=%s version=%s default=%s options=%s',
                     name, using, version, db_table_default, db_table_options)

        if table_name := db_table_options.get(version):
            meta.db_table = table_name
            logger.debug('Resolved table for %s using version %s → %s', name, version, table_name)
            return queryset

        if table_name := db_table_options.get(db_table_default):
            meta.db_table = table_name
            logger.debug('Resolved table for %s using default %s → %s', name, db_table_default, table_name)
            return queryset

        raise Exception(f'The model {name} does not have a table for version {version} using connection {using}')


class BaseModel(models.Model):
    objects = DynamicManager()

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        for field in self._meta.get_fields():
            if isinstance(field, models.CharField):
                val = str(getattr(self, field.name) or '')
                if val and field.max_length < len(val):
                    truncated = val[:field.max_length]
                    setattr(self, field.name, truncated)
                    logger.warning('Value field %s.%s was truncated to %s characters (%r:%r)',
                                   self.__class__.__name__, field.name, field.max_length, val, truncated)

        if not using: using = router.db_for_write(None)
        query = self.__class__.objects.using(using).filter(pk=self.pk)

        if self.pk and query.exists():
            fields = update_fields or [
                field.name for field in self._meta.fields
                if field.name != self._meta.pk.name and not isinstance(field, (models.AutoField, models.BigAutoField))
            ]
            values = {field: getattr(self, field) for field in fields}
            query.update(**values)
        else:
            super().save(force_insert, force_update, using, update_fields)

    class Meta:
        abstract = True


class DummyForwardManyToOneDescriptor(ForwardManyToOneDescriptor):
    def __get__(self, instance, cls=None):
        try:
            value = super().__get__(instance, cls)
        except self.field.remote_field.model.DoesNotExist:
            value = getattr(instance, self.field.get_attname())
        return value

    def __set__(self, instance, value):
        if value is not None and not isinstance(value, self.field.remote_field.model._meta.concrete_model):
            setattr(instance, self.field.get_attname(), value)
            instance._state.fields_cache.pop(self.field.name, None)
        else:
            return super().__set__(instance, value)


class DummyForeignKey(models.ForeignKey):
    forward_related_accessor_class = DummyForwardManyToOneDescriptor


def contribute_to_model(model, column, field):
    field.contribute_to_class(model, column)



from functools import lru_cache

from django.core.exceptions import FieldError
from django.db import router, connections
from django.db.models import Q

from isapilib.logging import logger


class ExecutionException(Exception):
    def __init__(self, exception, query, params, using):
        self.exception = exception
        self.params = params
        self.query = query
        self.using = using

    def __str__(self):
        return f'{self.using}: \n{self.query} ({self.params})\n{self.exception}'


class ExecutionQuery:
    def __init__(self, using):
        self._alias = using
        self.cursor = connections[using].cursor()

        self._query = None
        self._params = None
        self._log = True

        self.name_normalize = self.__class__.__name__.replace('Execution', '')

    def execute_query(self, query, params):
        try:
            self._query = query
            self._params = params

            if self._log: logger.debug('Execution %s, Query: "%s" %s (%s)', self.name_normalize, query, params,
                                       self._alias)
            self.cursor.execute(query, params)
        except Exception as e:
            raise ExecutionException(e, self._query, self._params, self._alias)

    def get_dataset(self):
        dataset = self.cursor.fetchall()
        return dataset

    def get_results(self):
        try:
            datasets: list = [self.get_dataset()]

            while self.cursor.nextset():
                datasets.append(self.get_dataset())

            if self._log: logger.debug('Results %s, Datasets: %s', self.name_normalize, datasets)
            return datasets
        except Exception as e:
            raise ExecutionException(e, self._query, self._params, self._alias)

    def execute(self, query, params):
        self.execute_query(query, params)
        try:
            return self.get_results()
        except Exception as e:
            if 'Previous SQL was not a query' in str(e):
                if self._log: logger.debug('Results %s, Datasets: %s', self.name_normalize, str(e))
                return None
            raise e

    def wrapper_query(self, query, _):
        return query

    def wrapper_params(self, params):
        if params is None: return []

        if not isinstance(params, list):
            return [params]

        return params

    def __call__(self, query, params=None, log=True):
        self._log = log

        params = self.wrapper_params(params)
        query = self.wrapper_query(query, params)
        return self.execute(query, params)


class ExecutionFunction(ExecutionQuery):
    def wrapper_query(self, function_name, params):
        return f"dbo.[{function_name}]({','.join('%s' for _ in params)})"


class ExecutionScalarFunction(ExecutionFunction):
    def wrapper_query(self, function_name, params):
        function_definition = super().wrapper_query(function_name, params)
        return f'SELECT {function_definition}'

    def get_results(self):
        results = super().get_results()
        return results[0][0][0]


class ExecutionTableFunction(ExecutionFunction):
    def __init__(self, using):
        super().__init__(using)
        self.fields = []

    def wrapper_query(self, function_name, params):
        function_definition = super().wrapper_query(function_name, params)

        fields_string = '*'
        if isinstance(self.fields, list):
            fields_string = ','.join(self.fields)

        return f'''SELECT {fields_string} FROM {function_definition}'''

    def __call__(self, query, params=None, fields=None, *args, **kwargs):
        self.fields = fields
        return super().__call__(query, params, *args, **kwargs)


class ExecutionStoredProcedure(ExecutionQuery):
    def __init__(self, using):
        super().__init__(using)
        self.only_output = True
        self.has_outputs = False

    def get_outputs(self, procedure_name):
        query = (
            "SELECT PARAMETER_NAME, DATA_TYPE, CHARACTER_MAXIMUM_LENGTH "
            "FROM information_schema.parameters "
            "WHERE SPECIFIC_NAME = %s AND PARAMETER_MODE = 'INOUT'"
        )
        self.cursor.execute(query, [procedure_name])
        results = self.cursor.fetchall()
        return 0 < len(results), results

    def wrapper_params(self, params):
        if isinstance(params, dict):
            return params

        return super().wrapper_params(params)

    def execute(self, query, params):
        if isinstance(params, dict): params = list(params.values())
        return super().execute(query, params)

    @staticmethod
    def get_parameters(params):
        if isinstance(params, list):
            return ','.join('%s' for _ in params)

        if isinstance(params, dict):
            return ','.join([f'@{i}=%s' for i in params.keys()])

        raise Exception('Invalida params type, excepted (list, dict)')

    def wrapper_query(self, procedure_name, params):
        def get_parameters_definition(name, kind, attr):
            if kind.lower() in ('nvarchar', 'varchar', 'char'):
                length = 'max' if attr == -1 else attr
                kind = f'{kind}({length})'

            return f'{name} {kind}'

        self.has_outputs, outputs = self.get_outputs(procedure_name)

        procedure_params = self.get_parameters(params)
        query = f"EXEC {procedure_name} {procedure_params}"

        if self.has_outputs:
            query = (
                f'''DECLARE {','.join([get_parameters_definition(*output) for output in outputs])}\n'''
                f'''{query}{',' if len(params) > 0 else ''}{','.join([f'{o[0]}={o[0]} OUTPUT' for o in outputs])}\n'''
                f'''SELECT 'OUTPUT',{','.join([o[0] for o in outputs])}'''
            )

        return f'SET NOCOUNT ON\n{query}'

    def get_results(self):
        datasets: list = super().get_results()

        last = datasets[-1]
        for row in last:
            if row[0] == 'OUTPUT':
                results = [row[1:] for row in last]

                if self.has_outputs:
                    datasets[-1] = results

                if self.only_output:
                    return results

        if self.only_output:
            return None

        return datasets

    def __call__(self, query, params=None, only_output=True, *args, **kwargs):
        self.only_output = only_output
        return super().__call__(query, params, *args, **kwargs)


def execute_query(*args, using=None, **kwargs):
    if not using: using = router.db_for_write(None)
    return ExecutionQuery(using)(*args, **kwargs)


def execute_fn(*args, using=None, **kwargs):
    if not using: using = router.db_for_write(None)
    return ExecutionScalarFunction(using)(*args, **kwargs)


def execute_table_fn(*args, using=None, **kwargs):
    if not using: using = router.db_for_write(None)
    return ExecutionTableFunction(using)(*args, **kwargs)


def execute_sp(*args, using=None, **kwargs):
    if not using: using = router.db_for_write(None)
    return ExecutionStoredProcedure(using)(*args, **kwargs)


def get_sucursal(mov='Servicio', sucursal=0):
    from isapilib.models import Sucursal
    if issubclass(type(sucursal), Sucursal): sucursal = getattr(sucursal, 'pk')

    if (mov in ['Venta Perdida', 'Dias', 'Reservar'] or 'Nota' in mov) and sucursal % 2 == 1:
        sucursal -= 1

    if mov in ['Cita Servicio'] and int(sucursal) % 2 == 0:
        sucursal += 1

    return Sucursal.objects.get(pk=sucursal).pk


def get_almacen(mov='Servicio', sucursal=0):
    from isapilib.models import Sucursal, Almacen
    if issubclass(type(sucursal), Sucursal): sucursal = getattr(sucursal, 'pk')

    if (mov in ['Venta Perdida', 'Hist Refacc', 'Reservar'] or 'Nota' in mov) and sucursal % 2 == 1:
        sucursal -= 1
        return Almacen.objects.get(Q(sucursal=sucursal), Q(almacen='R') | Q(almacen__istartswith='RS')).pk
    else:
        return Almacen.objects.get(sucursal=sucursal, almacen__istartswith='S').pk


def get_uen(modulo="VTAS", mov='Servicio', sucursal=0, concepto='Publico', using=None):
    from isapilib.models import Sucursal
    if issubclass(type(sucursal), Sucursal): sucursal = getattr(sucursal, 'pk')
    return execute_fn("fnCA_GeneraUENValida", [modulo, mov, sucursal, concepto], using=using)


def get_param_empresa(interfaz, clave, default=None, using=None):
    from isapilib.models import InterfacesPredefinidasDEmpresa as Data
    if not using: using = router.db_for_write(None)
    valor = Data.objects.using(using).filter(clave=clave, interfaces__interfaz=interfaz).first()
    return getattr(valor, 'valor_default', default)


def get_param_sucursal(sucursal, clave, default=None, using=None):
    from isapilib.models import Sucursal, ParametrosSucursal as Data
    if not using: using = router.db_for_write(None)
    if issubclass(type(sucursal), Sucursal): sucursal = sucursal.pk
    valor = Data.objects.using(using).filter(sucursal=sucursal, clave=clave).first()
    return getattr(valor, 'valor', default)


def verify_col(table, column, using=None) -> bool:
    try:
        query = f"SELECT 1 FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME = '{table}' AND COLUMN_NAME = '{column}'"
        result = execute_query(query, using)
        exists = result[0][0][0]
        return bool(exists)
    except IndexError:
        return False


@lru_cache
def get_utc_offset(using=None) -> int:
    result = execute_query('SELECT DATEPART(TZOFFSET, SYSDATETIMEOFFSET())', [], using=using)[0][0][0]
    return int(result)


def filter_queryset(queryset, params, field_mapping):
    for param in params:
        if param not in field_mapping:
            continue

        field_name = field_mapping[param]
        try:
            filters = {f'{field_name}__icontains': params[param]}
            queryset = queryset.filter(**filters)
        except FieldError:
            filters = {field_name: params[param]}
            queryset = queryset.filter(**filters)

    return queryset



import json
from datetime import timezone, datetime

from django.conf import settings

from isapilib.api.models import ApiLogs
from isapilib.core.get_functions import get_request_content_function, get_response_content_function
from isapilib.core.utilities import is_test
from isapilib.logging import logger

get_request_content = get_request_content_function()
get_response_content = get_response_content_function()


def insert_log(request, response, interfaz, tipo, time=0):
    force_insert = getattr(settings, 'FORCE_INSERT_LOGS', False)

    if is_test() and not force_insert:
        logger.debug("Skipping log insert in test mode (force_insert=%s)", force_insert)
        return

    try:
        log = ApiLogs()
        log.user_id = getattr(request.user, 'pk', None)
        log.tipo = str(tipo or '')
        log.header = str(json.dumps(dict(request.headers)))
        log.request = get_request_content(request)
        log.response = get_response_content(request, response)
        log.status = response.status_code
        log.url = request.build_absolute_uri()
        log.interfaz = str(interfaz or '')
        log.response_time = time
        log.fecharegistro = datetime.now(tz=timezone.utc)
        log.save(using='default')
    except Exception as e:
        logger.warning(f'Failed to save log: (%s) %s', type(e).__name__, e, exc_info=True)



from django.db import models

from isapilib.core.models import BaseModel, DummyForeignKey
from isapilib.core.utilities import get_default_user
from isapilib.external.utilities import execute_sp
from isapilib.logging import logger
from isapilib.mixin.venta import VentaMixin
from isapilib.mixin.vin import VinMixin


class Almacen(BaseModel):
    almacen = models.CharField(db_column='Almacen', primary_key=True, max_length=10)
    rama = models.CharField(db_column='Rama', max_length=20, blank=True, null=True)
    nombre = models.CharField(db_column='Nombre', max_length=100)
    direccion = models.CharField(db_column='Direccion', max_length=100, blank=True, null=True)
    categoria = models.CharField(db_column='Categoria', max_length=50, blank=True, null=True)
    tipo = models.CharField(db_column='Tipo', max_length=15, blank=True, null=True)
    sucursal = models.IntegerField(db_column='Sucursal')

    class Meta:
        managed = False
        db_table = 'Alm'


class Art(BaseModel):
    articulo = models.CharField(db_column='Articulo', primary_key=True, max_length=20)
    descripcion1 = models.CharField(db_column='Descripcion1', max_length=100, blank=True, null=True)
    descripcion2 = models.CharField(db_column='Descripcion2', max_length=255, blank=True, null=True)
    grupo = models.CharField(db_column='Grupo', max_length=50, blank=True, null=True)
    categoria = models.CharField(db_column='Categoria', max_length=50, blank=True, null=True)
    fabricante = models.CharField(db_column='Fabricante', max_length=50, blank=True, null=True)
    impuesto1 = models.FloatField(db_column='Impuesto1')
    impuesto2 = models.FloatField(db_column='Impuesto2', blank=True, null=True)
    impuesto3 = models.FloatField(db_column='Impuesto3', blank=True, null=True)
    tipo = models.CharField(db_column='Tipo', max_length=20)
    precio_lista = models.DecimalField(db_column='PrecioLista', max_digits=19, decimal_places=4, blank=True, null=True)
    estatus = models.CharField(db_column='Estatus', max_length=15, default='ALTA')
    usuario = models.CharField(db_column='Usuario', max_length=10, default=get_default_user)
    precio2 = models.DecimalField(db_column='Precio2', max_digits=19, decimal_places=4, blank=True, null=True)
    precio3 = models.DecimalField(db_column='Precio3', max_digits=19, decimal_places=4, blank=True, null=True)
    precio4 = models.DecimalField(db_column='Precio4', max_digits=19, decimal_places=4, blank=True, null=True)
    precio5 = models.DecimalField(db_column='Precio5', max_digits=19, decimal_places=4, blank=True, null=True)
    precio6 = models.DecimalField(db_column='Precio6', max_digits=19, decimal_places=4, blank=True, null=True)
    precio7 = models.DecimalField(db_column='Precio7', max_digits=19, decimal_places=4, blank=True, null=True)
    precio8 = models.DecimalField(db_column='Precio8', max_digits=19, decimal_places=4, blank=True, null=True)
    precio9 = models.DecimalField(db_column='Precio9', max_digits=19, decimal_places=4, blank=True, null=True)
    precio10 = models.DecimalField(db_column='Precio10', max_digits=19, decimal_places=4, blank=True, null=True)
    situacion = models.CharField(db_column='Situacion', max_length=50, blank=True, null=True)
    tipo_compra = models.CharField(db_column='TipoCompra', max_length=20, blank=True, null=True)
    retencion1 = models.FloatField(db_column='Retencion1', blank=True, null=True)
    retencion2 = models.FloatField(db_column='Retencion2', blank=True, null=True)
    retencion3 = models.FloatField(db_column='Retencion3', blank=True, null=True)
    modelo = models.CharField(db_column='Modelo', max_length=4, blank=True, null=True)
    direccion = models.CharField(db_column='Direccion', max_length=100, blank=True, null=True)
    direccion_numero = models.CharField(db_column='DireccionNumero', max_length=20, blank=True, null=True)
    direccion_numero_int = models.CharField(db_column='DireccionNumeroInt', max_length=20, blank=True, null=True)
    observaciones = models.CharField(db_column='Observaciones', max_length=100, blank=True, null=True)
    colonia = models.CharField(db_column='Colonia', max_length=100, blank=True, null=True)
    delegacion = models.CharField(db_column='Delegacion', max_length=100, blank=True, null=True)
    poblacion = models.CharField(db_column='Poblacion', max_length=100, blank=True, null=True)
    estado = models.CharField(db_column='Estado', max_length=30, blank=True, null=True)
    pais = models.CharField(db_column='Pais', max_length=30, blank=True, null=True)
    codigo_postal = models.CharField(db_column='CodigoPostal', max_length=15, blank=True, null=True)
    tipo_impuesto1 = models.CharField(db_column='TipoImpuesto1', max_length=10, blank=True, null=True)
    tipo_impuesto2 = models.CharField(db_column='TipoImpuesto2', max_length=10, blank=True, null=True)
    tipo_impuesto3 = models.CharField(db_column='TipoImpuesto3', max_length=10, blank=True, null=True)
    unidad_cantidad = models.FloatField(db_column='UnidadCantidad', blank=True, null=True)

    class Meta:
        managed = False
        db_table = 'Art'

    def __str__(self):
        return self.articulo


class ArtExistenciaNeta(BaseModel):
    articulo = models.CharField(db_column='Articulo', max_length=10, primary_key=True)
    empresa = models.CharField(db_column='Empresa', max_length=10)
    almacen = models.CharField(db_column='Almacen', max_length=10)
    moneda = models.CharField(db_column='Moneda', max_length=10)
    existencia = models.CharField(db_column='Existencia', max_length=10)

    class Meta:
        managed = False
        db_table = 'ArtExistenciaNeta'


class Empresa(BaseModel):
    empresa = models.CharField(db_column='Empresa', primary_key=True, max_length=5)
    nombre = models.CharField(db_column='Nombre', max_length=100, blank=True, null=True)
    grupo = models.CharField(db_column='Grupo', max_length=100, blank=True, null=True)
    direccion = models.CharField(db_column='Direccion', max_length=100, blank=True, null=True)
    direccion_numero = models.CharField(db_column='DireccionNumero', max_length=20, blank=True, null=True)
    direccion_numero_int = models.CharField(db_column='DireccionNumeroInt', max_length=20, blank=True, null=True)
    colonia = models.CharField(db_column='Colonia', max_length=100, blank=True, null=True)
    poblacion = models.CharField(db_column='Poblacion', max_length=30, blank=True, null=True)
    estado = models.CharField(db_column='Estado', max_length=30, blank=True, null=True)
    pais = models.CharField(db_column='Pais', max_length=30, blank=True, null=True)
    codigo_postal = models.CharField(db_column='CodigoPostal', max_length=15, blank=True, null=True)
    telefonos = models.CharField(db_column='Telefonos', max_length=100, blank=True, null=True)
    rfc = models.CharField(db_column='RFC', max_length=20, blank=True, null=True)
    tipo = models.CharField(db_column='Tipo', max_length=20, blank=True, null=True)

    class Meta:
        managed = False
        db_table = 'Empresa'


class Agente(BaseModel):
    agente = models.CharField(db_column='Agente', primary_key=True, unique=True, max_length=10)
    nombre = models.CharField(db_column='Nombre', max_length=100, blank=True, null=True)
    tipo = models.CharField(db_column='Tipo', max_length=15, blank=True, null=True)
    categoria = models.CharField(db_column='Categoria', max_length=50, blank=True, null=True)
    familia = models.CharField(db_column='Familia', max_length=50, blank=True, null=True)
    zona = models.CharField(db_column='Zona', max_length=30, blank=True, null=True)
    grupo = models.CharField(db_column='Grupo', max_length=50, blank=True, null=True)
    estatus = models.CharField(db_column='Estatus', max_length=15, default='ALTA')
    alta = models.DateTimeField(db_column='Alta', blank=True, null=True)
    baja = models.DateTimeField(db_column='Baja', blank=True, null=True)
    direccion = models.CharField(db_column='Direccion', max_length=100, blank=True, null=True)
    colonia = models.CharField(db_column='Colonia', max_length=255, blank=True, null=True)
    poblacion = models.CharField(db_column='Poblacion', max_length=30, blank=True, null=True)
    estado = models.CharField(db_column='Estado', max_length=30, blank=True, null=True)
    pais = models.CharField(db_column='Pais', max_length=30, blank=True, null=True)
    codigo_postal = models.CharField(db_column='CodigoPostal', max_length=15, blank=True, null=True)
    rfc = models.CharField(db_column='RFC', max_length=20, blank=True, null=True)
    curp = models.CharField(db_column='CURP', max_length=30, blank=True, null=True)
    email = models.CharField(db_column='eMail', max_length=50, blank=True, null=True)
    id_planta = models.CharField(db_column='IdPlanta', max_length=10, blank=True, null=True)
    personal_nombres = models.CharField(db_column='PersonalNombres', max_length=40)
    personal_apellido_paterno = models.CharField(db_column='PersonalApellidoPaterno', max_length=30)
    personal_apellido_materno = models.CharField(db_column='PersonalApellidoMaterno', max_length=30)

    class Meta:
        managed = False
        db_table = 'Agente'


class Cte(BaseModel):
    cliente = models.CharField(db_column='Cliente', primary_key=True, max_length=10)
    nombre = models.CharField(db_column='Nombre', max_length=254, blank=True, null=True)
    nombre_corto = models.CharField(db_column='NombreCorto', max_length=20, blank=True, null=True)
    direccion = models.CharField(db_column='Direccion', max_length=100, blank=True, null=True)
    direccion_numero = models.CharField(db_column='DireccionNumero', max_length=50, blank=True, null=True)
    observaciones = models.CharField(db_column='Observaciones', max_length=100, blank=True, null=True)
    delegacion = models.CharField(db_column='Delegacion', max_length=100, blank=True, null=True)
    colonia = models.CharField(db_column='Colonia', max_length=100, blank=True, null=True)
    poblacion = models.CharField(db_column='Poblacion', max_length=100, blank=True, null=True)
    estado = models.CharField(db_column='Estado', max_length=30, blank=True, null=True)
    pais = models.CharField(db_column='Pais', max_length=100, blank=True, null=True)
    zona = models.CharField(db_column='Zona', max_length=30, blank=True, null=True)
    codigo_postal = models.CharField(db_column='CodigoPostal', max_length=15, blank=True, null=True)
    rfc = models.CharField(db_column='RFC', max_length=15, blank=True, null=True)
    curp = models.CharField(db_column='CURP', max_length=30, blank=True, null=True)
    telefonos = models.CharField(db_column='Telefonos', max_length=100, blank=True, null=True)
    telefonos_lada = models.CharField(db_column='TelefonosLada', max_length=6, blank=True, null=True)
    contacto1 = models.CharField(db_column='Contacto1', max_length=50, blank=True, null=True)
    contacto2 = models.CharField(db_column='Contacto2', max_length=50, blank=True, null=True)
    extencion1 = models.CharField(db_column='Extencion1', max_length=10, blank=True, null=True)
    extencion2 = models.CharField(db_column='Extencion2', max_length=10, blank=True, null=True)
    email = models.CharField(db_column='eMail', max_length=50, blank=True, null=True)
    email1 = models.CharField(db_column='eMail1', max_length=50, blank=True, null=True)
    email2 = models.CharField(db_column='eMail2', max_length=50, blank=True, null=True)
    categoria = models.CharField(db_column='Categoria', max_length=50, blank=True, null=True)
    tipo = models.CharField(db_column='Tipo', max_length=15, blank=True, null=True)
    situacion = models.CharField(db_column='Situacion', max_length=50, blank=True, null=True)
    agente = DummyForeignKey(Agente, db_column='agente', to_field='agente', on_delete=models.PROTECT,
                             related_name='clientes', blank=True, null=True)
    agente_servicio = models.CharField(db_column='AgenteServicio', max_length=10, blank=True, null=True)
    estatus = models.CharField(db_column='Estatus', max_length=15, default='ALTA')
    ultimo_cambio = models.DateTimeField(db_column='UltimoCambio', blank=True, null=True)
    descripcion1 = models.CharField(db_column='Descripcion1', max_length=50, blank=True, null=True)
    descripcion2 = models.CharField(db_column='Descripcion2', max_length=50, blank=True, null=True)
    descripcion3 = models.CharField(db_column='Descripcion3', max_length=50, blank=True, null=True)
    descripcion4 = models.CharField(db_column='Descripcion4', max_length=50, blank=True, null=True)
    descripcion5 = models.CharField(db_column='Descripcion5', max_length=50, blank=True, null=True)
    descripcion6 = models.CharField(db_column='Descripcion6', max_length=50, blank=True, null=True)
    descripcion7 = models.CharField(db_column='Descripcion7', max_length=50, blank=True, null=True)
    descripcion8 = models.CharField(db_column='Descripcion8', max_length=50, blank=True, null=True)
    descripcion9 = models.CharField(db_column='Descripcion9', max_length=50, blank=True, null=True)
    descripcion10 = models.CharField(db_column='Descripcion10', max_length=50, blank=True, null=True)
    descripcion11 = models.CharField(db_column='Descripcion11', max_length=50, blank=True, null=True)
    descripcion12 = models.CharField(db_column='Descripcion12', max_length=50, blank=True, null=True)
    descripcion13 = models.CharField(db_column='Descripcion13', max_length=50, blank=True, null=True)
    descripcion14 = models.CharField(db_column='Descripcion14', max_length=50, blank=True, null=True)
    descripcion15 = models.CharField(db_column='Descripcion15', max_length=50, blank=True, null=True)
    descripcion16 = models.CharField(db_column='Descripcion16', max_length=50, blank=True, null=True)
    descripcion17 = models.CharField(db_column='Descripcion17', max_length=50, blank=True, null=True)
    descripcion18 = models.CharField(db_column='Descripcion18', max_length=50, blank=True, null=True)
    descripcion19 = models.CharField(db_column='Descripcion19', max_length=50, blank=True, null=True)
    descripcion20 = models.CharField(db_column='Descripcion20', max_length=50, blank=True, null=True)
    personal_nombres = models.CharField(db_column='PersonalNombres', max_length=50, blank=True, null=True)
    personal_apellido_paterno = models.CharField(db_column='PersonalApellidoPaterno', max_length=50, blank=True,
                                                 null=True)
    personal_apellido_materno = models.CharField(db_column='PersonalApellidoMaterno', max_length=50, blank=True,
                                                 null=True)
    personal_direccion = models.CharField(db_column='PersonalDireccion', max_length=100, blank=True, null=True)
    personal_entrecalles = models.CharField(db_column='PersonalEntreCalles', max_length=100, blank=True, null=True)
    personal_plano = models.CharField(db_column='PersonalPlano', max_length=15, blank=True, null=True)
    personal_delegacion = models.CharField(db_column='PersonalDelegacion', max_length=100, blank=True, null=True)
    personal_colonia = models.CharField(db_column='PersonalColonia', max_length=100, blank=True, null=True)
    personal_poblacion = models.CharField(db_column='PersonalPoblacion', max_length=100, blank=True, null=True)
    personal_estado = models.CharField(db_column='PersonalEstado', max_length=30, blank=True, null=True)
    personal_pais = models.CharField(db_column='PersonalPais', max_length=30, blank=True, null=True)
    personal_zona = models.CharField(db_column='PersonalZona', max_length=30, blank=True, null=True)
    personal_codigo_postal = models.CharField(db_column='PersonalCodigoPostal', max_length=15, blank=True, null=True)
    personal_telefonos = models.CharField(db_column='PersonalTelefonos', max_length=100, blank=True, null=True)
    personal_telefonos_lada = models.CharField(db_column='PersonalTelefonosLada', max_length=6, blank=True, null=True)
    personal_telefono_movil = models.CharField(db_column='PersonalTelefonoMovil', max_length=30, blank=True, null=True)
    personal_sms = models.BooleanField(db_column='PersonalSMS', blank=True, null=True)
    fecha_nacimiento = models.DateTimeField(db_column='FechaNacimiento', blank=True, null=True)
    sexo = models.CharField(db_column='Sexo', max_length=20, blank=True, null=True)
    fecha1 = models.DateTimeField(db_column='Fecha1', blank=True, null=True)
    fecha2 = models.DateTimeField(db_column='Fecha2', blank=True, null=True)
    fecha3 = models.DateTimeField(db_column='Fecha3', blank=True, null=True)
    fecha4 = models.DateTimeField(db_column='Fecha4', blank=True, null=True)
    fecha5 = models.DateTimeField(db_column='Fecha5', blank=True, null=True)
    usuario = models.CharField(db_column='Usuario', max_length=10, default=get_default_user)
    fiscal_regimen = models.CharField(db_column='FiscalRegimen', max_length=30, blank=True, null=True)
    contactar = models.CharField(db_column='Contactar', max_length=30, blank=True, null=True)

    def save(self, *args, **kwargs):
        if not self.pk:
            usuario = get_default_user()

            self.cliente = execute_sp('spConsecutivo', ['CTE', 0])[0][0]
            self.colonia = self.colonia or 'NA'
            self.rfc = self.rfc or 'XAXX010101000'
            self.direccion = self.direccion or 'NA'
            self.direccion_numero = self.direccion_numero or 'XAXX010101000'
            self.estatus = self.estatus or 'ALTA'
            self.tipo = self.tipo or 'Cliente'
            self.estado = self.estado or 'NA'
            self.pais = self.pais or 'NA'
            self.usuario = usuario

            logger.debug("Created a new client %s", self.pk)

        return super().save(*args, **kwargs)

    class Meta:
        managed = False
        db_table = 'Cte'


class Sucursal(BaseModel):
    sucursal = models.IntegerField(db_column='Sucursal', primary_key=True)
    nombre = models.CharField(db_column='Nombre', max_length=100, blank=True, null=True)
    prefijo = models.CharField(db_column='Prefijo', max_length=5, blank=True, null=True)
    estatus = models.CharField(db_column='Estatus', max_length=15, default='ALTA')
    rfc = models.CharField(db_column='RFC', max_length=20, blank=True, null=True)
    alta = models.DateTimeField(db_column='Alta', blank=True, null=True)
    almacen_principal = models.CharField(db_column='AlmacenPrincipal', max_length=10, blank=True, null=True)
    cliente = models.CharField(db_column='Cliente', max_length=10, blank=True, null=True)
    categoria = models.CharField(db_column='Categoria', max_length=50, blank=True, null=True)
    ip = models.CharField(db_column='IP', max_length=20, blank=True, null=True)
    fiscal_regimen = models.CharField(db_column='FiscalRegimen', max_length=30, blank=True, null=True)
    version = models.CharField(db_column='Version', max_length=25, blank=True, null=True)
    gwmbac = models.CharField(db_column='GWMBAC', max_length=11, blank=True, null=True)

    class Meta:
        managed = False
        db_table = 'Sucursal'


class Compra(BaseModel):
    id = models.AutoField(db_column='ID', primary_key=True, unique=True)
    empresa = DummyForeignKey(Empresa, db_column='Empresa', to_field='empresa', on_delete=models.PROTECT,
                              related_name='compras')
    sucursal = DummyForeignKey(Sucursal, db_column='Sucursal', to_field='sucursal', on_delete=models.PROTECT,
                               related_name='compras')
    almacen = DummyForeignKey(Almacen, db_column='Almacen', to_field='almacen', on_delete=models.PROTECT,
                              related_name='compras', blank=True, null=True)
    importe = models.DecimalField(db_column='Importe', max_digits=19, decimal_places=4, blank=True, null=True)
    impuestos = models.DecimalField(db_column='Impuestos', max_digits=19, decimal_places=4, blank=True, null=True)

    class Meta:
        managed = False
        db_table = 'Compra'


class Vin(BaseModel, VinMixin):
    vin = models.CharField(db_column='VIN', primary_key=True, max_length=20)
    articulo = DummyForeignKey(Art, db_column='Articulo', to_field='articulo', on_delete=models.PROTECT,
                               related_name='vins', blank=True, null=True)
    km = models.IntegerField(db_column='Km', blank=True, null=True)
    motor = models.CharField(db_column='Motor', max_length=20, blank=True, null=True)
    fecha = models.DateTimeField(db_column='Fecha', blank=True, null=True)
    cliente = DummyForeignKey(Cte, db_column='Cliente', to_field='cliente', on_delete=models.PROTECT,
                              related_name='vins')
    conductor = models.CharField(db_column='Conductor', max_length=10, blank=True, null=True)
    alta = models.DateTimeField(db_column='Alta', blank=True, null=True)
    empresa = DummyForeignKey(Empresa, db_column='Empresa', to_field='empresa', on_delete=models.PROTECT,
                              related_name='vins')
    placas = models.CharField(db_column='Placas', max_length=20, blank=True, null=True)
    garantia_vencimiento = models.DateTimeField(db_column='GarantiaVencimiento', blank=True, null=True)
    registro = models.CharField(db_column='Registro', max_length=20, blank=True, null=True)
    fecha_carta_credito = models.DateTimeField(db_column='FechaCartaCredito', blank=True, null=True)
    fecha_factura = models.DateTimeField(db_column='FechaFactura', blank=True, null=True)
    fecha_ultimo_servicio = models.DateTimeField(db_column='FechaUltimoServicio', blank=True, null=True)
    fecha_siguiente_servicio = models.DateTimeField(db_column='FechaSiguienteServicio', blank=True, null=True)
    costo = models.FloatField(db_column='Costo', blank=True, null=True)
    modelo = models.CharField(db_column='Modelo', max_length=4, blank=True, null=True)
    tipo_compra = models.CharField(db_column='TipoCompra', max_length=1, blank=True, null=True)
    folio_factura_compra = models.CharField(db_column='FolioFacturaCompra', max_length=15, blank=True, null=True)
    fecha_factura_compra = models.DateTimeField(db_column='FechaFacturaCompra', blank=True, null=True)
    descripcion1 = models.CharField(db_column='Descripcion1', max_length=38, blank=True, null=True)
    descripcion2 = models.CharField(db_column='Descripcion2', max_length=38, blank=True, null=True)
    color_exterior = models.CharField(db_column='ColorExterior', max_length=10, blank=True, null=True)
    color_exterior_descripcion = models.CharField(db_column='ColorExteriorDescripcion', max_length=50, blank=True,
                                                  null=True)
    color_interior = models.CharField(db_column='ColorInterior', max_length=10, blank=True, null=True)
    color_interior_descripcion = models.CharField(db_column='ColorInteriorDescripcion', max_length=50, blank=True,
                                                  null=True)
    fecha_pago = models.DateTimeField(db_column='FechaPago', blank=True, null=True)
    venta_id = models.IntegerField(db_column='VentaID', blank=True, null=True)
    estatus = models.CharField(db_column='Estatus', max_length=15, default='ALTA')
    situacion = models.CharField(db_column='Situacion', max_length=50, blank=True, null=True)
    situacion_fecha = models.DateTimeField(db_column='SituacionFecha', blank=True, null=True)
    agente = DummyForeignKey(Agente, db_column='agente', to_field='agente', on_delete=models.PROTECT,
                             related_name='vins', blank=True, null=True)
    tipo_venta = models.CharField(db_column='TipoVenta', max_length=1, blank=True, null=True)
    kilometraje_inicial = models.IntegerField(db_column='KilometrajeInicial', blank=True, null=True)
    cilindros = models.IntegerField(db_column='Cilindros', blank=True, null=True)
    puertas = models.CharField(db_column='Puertas', max_length=30, blank=True, null=True)
    pasajeros = models.IntegerField(db_column='Pasajeros', blank=True, null=True)
    capacidad_carga = models.IntegerField(db_column='CapacidadCarga', blank=True, null=True)
    combustible = models.CharField(db_column='Combustible', max_length=20, blank=True, null=True)
    primera_llamada = models.DateTimeField(db_column='PrimeraLlamada', blank=True, null=True)
    comentarios_primera_llamada = models.CharField(db_column='ComentariosPrimeraLLamada', max_length=1000, blank=True,
                                                   null=True)
    segunda_llamada = models.DateTimeField(db_column='SegundaLlamada', blank=True, null=True)
    comentarios_segunda_llamada = models.CharField(db_column='ComentariosSegundaLLamada', max_length=1000, blank=True,
                                                   null=True)
    tercera_llamada = models.DateTimeField(db_column='TerceraLlamada', blank=True, null=True)
    comentarios_tercera_llamada = models.CharField(db_column='ComentariosTerceraLLamada', max_length=1000, blank=True,
                                                   null=True)
    demo = models.BooleanField(db_column='VehiculoDemo', blank=True, null=True)
    compra = DummyForeignKey(Compra, db_column='CompraID', to_field='id', on_delete=models.PROTECT,
                             related_name='vins', blank=True, null=True)

    class Meta:
        managed = False
        db_table = 'VIN'

    def __str__(self):
        return self.vin


class VinTipoAccesorio(BaseModel):
    tipo = models.CharField(db_column='Tipo', primary_key=True, max_length=50)
    afecta_costo = models.BooleanField(db_column='AfectaCosto')
    afecta_precio = models.BooleanField(db_column='AfectaPrecio')
    afecta_gasto = models.BooleanField(db_column='AfectaGasto')

    class Meta:
        managed = False
        db_table = 'VinTipoAccesorio'


class SerieLote(BaseModel):
    empresa = DummyForeignKey(Empresa, db_column='Empresa', to_field='empresa', on_delete=models.PROTECT,
                              related_name='series_lote')
    sucursal = DummyForeignKey(Sucursal, db_column='Sucursal', to_field='sucursal', on_delete=models.PROTECT,
                               related_name='series_lote')
    DummyForeignKey(Almacen, db_column='Almacen', to_field='almacen', on_delete=models.PROTECT,
                    related_name='series_lote')
    serie_lote = models.CharField(db_column='SerieLote', max_length=50)
    existencia = models.FloatField(db_column='Existencia', blank=True, null=True)

    class Meta:
        managed = False
        db_table = 'SerieLote'


class VinAccesorio(BaseModel):
    vin = DummyForeignKey(Vin, db_column='VIN', to_field='vin', on_delete=models.PROTECT,
                          related_name='accessories')
    tipo = DummyForeignKey(VinTipoAccesorio, db_column='Tipo', to_field='tipo', on_delete=models.PROTECT,
                           related_name='accessories', blank=True, null=True)
    descripcion = models.CharField(db_column='Descripcion', max_length=100, blank=True, null=True)
    precio_contado = models.DecimalField(db_column='PrecioContado', max_digits=19,
                                         decimal_places=4, blank=True, null=True)

    class Meta:
        managed = False
        db_table = 'VinAccesorio'


class Usuario(BaseModel):
    usuario = models.CharField(db_column='Usuario', primary_key=True, max_length=10)
    nombre = models.CharField(db_column='Nombre', max_length=100)
    sucursal = models.IntegerField(db_column='Sucursal')
    def_agente = DummyForeignKey(Agente, db_column='DefAgente', to_field='agente', on_delete=models.PROTECT,
                                 related_name='usuarios', blank=True, null=True)
    email = models.CharField(db_column='eMail', max_length=50)
    estatus = models.CharField(db_column='Estatus', max_length=15, default='ALTA')

    class Meta:
        managed = False
        db_table = 'Usuario'


class Venta(BaseModel, VentaMixin):
    id = models.AutoField(db_column='ID', primary_key=True, unique=True)
    agente = DummyForeignKey(Agente, db_column='Agente', to_field='agente', on_delete=models.PROTECT,
                             related_name='ventas', blank=True, null=True)
    empresa = DummyForeignKey(Empresa, db_column='Empresa', to_field='empresa', on_delete=models.PROTECT,
                              related_name='ventas')
    mov = models.CharField(db_column='Mov', max_length=20)
    mov_id = models.CharField(db_column='MovID', max_length=20, blank=True, null=True)
    fecha_emision = models.DateTimeField(db_column='FechaEmision', blank=True, null=True)
    ultimo_cambio = models.DateTimeField(db_column='UltimoCambio', blank=True, null=True)
    concepto = models.CharField(db_column='Concepto', default='Publico', max_length=50, blank=True, null=True)
    uen = models.IntegerField(db_column='UEN', blank=True, null=True)
    moneda = models.CharField(db_column='Moneda', max_length=10, default='Pesos')
    tipo_cambio = models.FloatField(db_column='TipoCambio', default=1, blank=True, null=True)
    usuario = DummyForeignKey(Usuario, db_column='Usuario', to_field='usuario', on_delete=models.PROTECT,
                              related_name='ventas', default=get_default_user)
    referencia = models.CharField(db_column='Referencia', max_length=50, blank=True, null=True)
    observaciones = models.CharField(db_column='Observaciones', max_length=100, blank=True, null=True)
    estatus = models.CharField(db_column='Estatus', default='SINAFECTAR', max_length=15, blank=True, null=True)
    situacion = models.CharField(db_column='Situacion', max_length=50, blank=True, null=True)
    situacion_fecha = models.DateTimeField(db_column='SituacionFecha', blank=True, null=True)
    situacion_usuario = models.CharField(db_column='SituacionUsuario', max_length=10, blank=True, null=True)
    situacion_nota = models.CharField(db_column='SituacionNota', max_length=100, blank=True, null=True)
    cliente = DummyForeignKey(Cte, db_column='Cliente', to_field='cliente', on_delete=models.PROTECT,
                              related_name='ventas')
    almacen = DummyForeignKey(Almacen, db_column='Almacen', to_field='almacen', on_delete=models.PROTECT,
                              related_name='ventas')
    fecha_requerida = models.DateTimeField(db_column='FechaRequerida', blank=True, null=True)
    hora_requerida = models.CharField(db_column='HoraRequerida', max_length=5, blank=True, null=True)
    condicion = models.CharField(db_column='Condicion', max_length=50, blank=True, null=True)
    servicio_tipo = models.CharField(db_column='ServicioTipo', max_length=50, blank=True, null=True)
    servicio_articulo = DummyForeignKey(Art, db_column='ServicioArticulo', to_field='articulo',
                                        on_delete=models.PROTECT,
                                        related_name='servicios', blank=True, null=True)
    servicio_serie = DummyForeignKey(Vin, db_column='ServicioSerie', to_field='vin',
                                     on_delete=models.PROTECT,
                                     related_name='ventas')
    servicio_contrato = models.CharField(db_column='ServicioContrato', max_length=20, blank=True, null=True)
    servicio_contrato_id = models.CharField(db_column='ServicioContratoID', max_length=20, blank=True, null=True)
    servicio_contrato_tipo = models.CharField(db_column='ServicioContratoTipo', max_length=50, blank=True, null=True)
    servicio_descripcion = models.CharField(db_column='ServicioDescripcion', max_length=100, blank=True, null=True)
    servicio_fecha = models.DateTimeField(db_column='ServicioFecha', blank=True, null=True, auto_now=True)
    servicio_flotilla = models.BooleanField(db_column='ServicioFlotilla', blank=True, null=True)
    servicio_rampa = models.BooleanField(db_column='ServicioRampa', blank=True, null=True)
    servicio_identificador = models.CharField(db_column='ServicioIdentificador', max_length=20, blank=True, null=True)
    servicio_placas = models.CharField(db_column='ServicioPlacas', max_length=20, blank=True, null=True)
    servicio_kms = models.IntegerField(db_column='ServicioKms', blank=True, null=True)
    servicio_tipo_orden = models.CharField(db_column='ServicioTipoOrden', default='Publico', max_length=20, blank=True,
                                           null=True)
    servicio_tipo_operacion = models.CharField(db_column='ServicioTipoOperacion', default='Publico', max_length=50,
                                               blank=True, null=True)
    servicio_siniestro = models.CharField(db_column='ServicioSiniestro', max_length=20, blank=True, null=True)
    servicio_deducible_importe = models.DecimalField(db_column='ServicioDeducibleImporte', max_digits=19,
                                                     decimal_places=4, blank=True, null=True)
    servicio_numero = models.FloatField(db_column='ServicioNumero', default=1, blank=True, null=True)
    servicio_numero_economico = models.CharField(db_column='ServicioNumeroEconomico', max_length=20, blank=True,
                                                 null=True)
    servicio_aseguradora = models.CharField(db_column='ServicioAseguradora', max_length=10, blank=True, null=True)
    servicio_puntual = models.BooleanField(db_column='ServicioPuntual', blank=True, null=True)
    servicio_poliza = models.CharField(db_column='ServicioPoliza', max_length=20, blank=True, null=True)
    origen = models.CharField(db_column='Origen', max_length=20, blank=True, null=True)
    origen_id = models.CharField(db_column='OrigenID', max_length=20, blank=True, null=True)
    servicio_modelo = models.CharField(db_column='ServicioModelo', max_length=4, blank=True, null=True)
    sucursal = DummyForeignKey(Sucursal, db_column='Sucursal', to_field='sucursal', on_delete=models.PROTECT,
                               related_name='ventas')
    sucursal_destino = models.IntegerField(db_column='SucursalDestino', blank=True, null=True)
    comentarios = models.TextField(db_column='Comentarios', blank=True, null=True)
    fecha_entrega = models.DateTimeField(db_column='FechaEntrega', blank=True, null=True)
    hora_recepcion = models.CharField(db_column='HoraRecepcion', max_length=5, blank=True, null=True)
    lista_precios_esp = models.CharField(db_column='ListaPreciosEsp', max_length=20, blank=True, null=True,
                                         default='Precio Publico')
    endosar_a = models.CharField(db_column='EndosarA', max_length=10, blank=True, null=True)
    forma_pago_tipo = models.CharField(db_column='FormaPagoTipo', max_length=50, blank=True, null=True)
    causa = models.CharField(db_column='Causa', max_length=50, blank=True, null=True)

    class Meta:
        managed = False
        db_table = 'Venta'


class VentaD(BaseModel):
    venta = DummyForeignKey(Venta, db_column='ID', to_field='id', on_delete=models.PROTECT,
                            related_name='details')
    renglon = models.FloatField(db_column='Renglon')
    renglon_sub = models.IntegerField(db_column='RenglonSub')
    renglon_id = models.IntegerField(db_column='RenglonID', blank=True, null=True)
    renglon_tipo = models.CharField(db_column='RenglonTipo', max_length=1, blank=True, null=True)
    cantidad = models.FloatField(db_column='Cantidad', blank=True, null=True)
    almacen = DummyForeignKey(Almacen, db_column='Almacen', to_field='almacen', on_delete=models.PROTECT,
                              related_name='details', blank=True, null=True)
    articulo = DummyForeignKey(Art, db_column='Articulo', to_field='articulo', on_delete=models.PROTECT,
                               related_name='details')
    precio = models.FloatField(db_column='Precio', blank=True, null=True)
    precio_sugerido = models.FloatField(db_column='PrecioSugerido', blank=True, null=True)
    impuesto1 = models.FloatField(db_column='Impuesto1', blank=True, null=True)
    impuesto2 = models.FloatField(db_column='Impuesto2', blank=True, null=True)
    impuesto3 = models.FloatField(db_column='Impuesto3', blank=True, null=True)
    descripcion_extra = models.CharField(db_column='DescripcionExtra', max_length=100, blank=True, null=True)
    costo = models.DecimalField(db_column='Costo', max_digits=19, decimal_places=4, blank=True, null=True)
    aplica = models.CharField(db_column='Aplica', max_length=20, blank=True, null=True)
    aplica_id = models.CharField(db_column='AplicaID', max_length=20, blank=True, null=True)
    unidad = models.CharField(db_column='Unidad', max_length=50, blank=True, null=True)
    fecha_requerida = models.DateTimeField(db_column='FechaRequerida', blank=True, null=True)
    hora_requerida = models.CharField(db_column='HoraRequerida', max_length=5, blank=True, null=True)
    agente = DummyForeignKey(Agente, db_column='Agente', to_field='agente', on_delete=models.PROTECT,
                             related_name='details', blank=True, null=True)
    departamento = models.IntegerField(db_column='Departamento', blank=True, null=True)
    sucursal = DummyForeignKey(Sucursal, db_column='Sucursal', to_field='sucursal', on_delete=models.PROTECT,
                               related_name='details')
    sucursal_origen = DummyForeignKey(Sucursal, db_column='SucursalOrigen', to_field='sucursal',
                                      on_delete=models.PROTECT,
                                      related_name='details_origen', blank=True, null=True)
    uen = models.IntegerField(db_column='UEN', blank=True, null=True)
    precio_lista = models.DecimalField(db_column='PrecioLista', max_digits=19, decimal_places=4, blank=True, null=True)
    tipo_impuesto1 = models.CharField(db_column='TipoImpuesto1', max_length=10, blank=True, null=True)
    tipo_impuesto2 = models.CharField(db_column='TipoImpuesto2', max_length=10, blank=True, null=True)
    tipo_impuesto3 = models.CharField(db_column='TipoImpuesto3', max_length=10, blank=True, null=True)
    retencion1 = models.FloatField(db_column='Retencion1', blank=True, null=True)
    retencion2 = models.FloatField(db_column='Retencion2', blank=True, null=True)
    retencion3 = models.FloatField(db_column='Retencion3', blank=True, null=True)
    tipo_retencion1 = models.CharField(db_column='TipoRetencion1', max_length=10, blank=True, null=True)
    tipo_retencion2 = models.CharField(db_column='TipoRetencion2', max_length=10, blank=True, null=True)
    tipo_retencion3 = models.CharField(db_column='TipoRetencion3', max_length=10, blank=True, null=True)
    comentarios = models.CharField(db_column='Comentarios', max_length=250, blank=True, null=True)
    servicio_tipo_orden = models.CharField(db_column='ServicioTipoOrden', max_length=20, blank=True, null=True)
    articulo_actual = models.CharField(db_column='ArticuloActual', max_length=20, blank=True, null=True)
    ut = models.FloatField(db_column='UT', blank=True, null=True)
    cc_tiempo_tab = models.FloatField(db_column='CCTiempoTab', blank=True, null=True)
    paquete = models.IntegerField(db_column='Paquete', blank=True, null=True)
    descuento_linea = models.FloatField(db_column='DescuentoLinea', blank=True, null=True)
    descuento_importe = models.DecimalField(db_column='DescuentoImporte', blank=True, max_digits=19, decimal_places=4,
                                            null=True)
    cantidad_pendiente = models.FloatField(db_column='CantidadPendiente', blank=True, null=True)

    class Meta:
        managed = False
        db_table = 'VentaD'
        unique_together = (('id', 'renglon', 'renglon_sub'),)


class CaVenta(BaseModel):
    venta = DummyForeignKey(Venta, db_column='VentaId', to_field='id', on_delete=models.PROTECT, related_name='ca_venta')
    id_prospecto = models.CharField(db_column='idProspecto', max_length=200)

    class Meta:
        managed = False
        db_table = 'CA_Venta'


class CuentasPorCobrar(BaseModel):
    mov = models.CharField(db_column='Mov', max_length=20)
    mov_id = models.CharField(db_column='MovID', max_length=20, blank=True, null=True)

    class Meta:
        managed = False
        db_table = 'Cxc'


class TipoOrdenOperacion(BaseModel):
    interfaz = models.CharField(db_column='Interfaz', max_length=100)
    operacion_planta = models.CharField(db_column='OrdenOperacionPlanta', max_length=50)
    operacion_intelisis = models.CharField(db_column='OrdenOperacionIntelisis', max_length=50)

    class Meta:
        managed = False
        db_table = 'CA_MapeoTipoOrdenOperacion'
        unique_together = (('interfaz', 'operacion_intelisis'),)


class MaestroTipoOrdenOperacion(BaseModel):
    marca = models.CharField(db_column='Marca', max_length=10)
    interfaz = models.CharField(db_column='Interfaz', max_length=100)
    tipo_orden_operacion = models.CharField(db_column='TipoOrdenOperacion', max_length=50)
    valor = models.CharField(db_column='Valor', max_length=50)

    class Meta:
        managed = False
        db_table = 'CA_MaestroTipoOrdenOperacionporInterfaz'


class MensajeLista(BaseModel):
    mensaje = models.IntegerField(primary_key=True)
    descripcion = models.CharField(max_length=255, blank=True, null=True)
    tipo = models.CharField(max_length=50, default='ERROR', blank=True, null=True)
    ie = models.BooleanField(default=0, blank=True, null=True)

    class Meta:
        managed = False
        db_table = 'MensajeLista'


class VentaTraspasarArticulos(BaseModel):
    venta = models.IntegerField(db_column='ID')
    estacion = models.IntegerField(db_column='Estacion')
    rid = models.AutoField(db_column='RID', primary_key=True, unique=True)
    codigo = models.CharField(db_column='Codigo', max_length=30, blank=True, null=True)
    articulo = DummyForeignKey(Art, db_column='Articulo', to_field='articulo', on_delete=models.PROTECT,
                               related_name='traspasos', blank=True, null=True)
    cantidad = models.FloatField(db_column='Cantidad', blank=True, null=True)
    precio = models.FloatField(db_column='Precio', blank=True, null=True)
    costo = models.FloatField(db_column='Costo', blank=True, null=True)
    accion = models.CharField(db_column='Accion', max_length=20, blank=True, null=True)
    referencia = models.CharField(db_column='Referencia', max_length=50, blank=True, null=True)
    aplica_mov = models.CharField(db_column='AplicaMov', max_length=20, blank=True, null=True)
    aplica_mov_id = models.CharField(db_column='AplicaMovID', max_length=20, blank=True, null=True)
    sin_validar_reservado = models.BooleanField(db_column='SinValidarReservado', default=False)
    paquete = models.IntegerField(db_column='Paquete', blank=True, null=True)
    usuario = models.CharField(db_column='Usuario', max_length=20, default=get_default_user)
    contrasena = models.CharField(db_column='Contrasena', max_length=40, blank=True, null=True)
    descuento_linea = models.DecimalField(db_column='DescuentoLinea', max_digits=19, decimal_places=4, blank=True,
                                          null=True)
    descuento_importe = models.DecimalField(db_column='DescuentoImporte', max_digits=19, decimal_places=4, blank=True,
                                            null=True)
    renglon = models.FloatField(db_column='Renglon', blank=True, null=True)
    origen = models.CharField(max_length=100, blank=True, null=True)

    class Meta:
        managed = False
        db_table = 'VentaTraspasarArticulos'
        unique_together = (('venta', 'estacion', 'rid'),)


class FormaPagoTipo(BaseModel):
    tipo = models.CharField(primary_key=True, max_length=50)
    sobre_precio = models.FloatField(db_column='SobrePrecio', blank=True, null=True)

    class Meta:
        managed = False
        db_table = 'FormaPagoTipo'


class ListaPreciosD(BaseModel):
    lista = models.CharField(db_column='Lista', max_length=20)
    moneda = models.CharField(db_column='Moneda', max_length=10)
    articulo = DummyForeignKey(Art, db_column='Articulo', to_field='articulo', on_delete=models.PROTECT,
                               related_name='lista_precios_details')
    precio = models.DecimalField(db_column='Precio', max_digits=18, decimal_places=4, null=True, blank=True)
    codigo_cliente = models.CharField(db_column='CodigoCliente', max_length=20, null=True, blank=True)
    margen = models.FloatField(db_column='Margen', null=True, blank=True)
    region = models.CharField(db_column='Region', max_length=50, null=True, blank=True)

    class Meta:
        managed = False
        db_table = 'ListaPreciosD'
        unique_together = (('lista', 'moneda', 'articulo'),)


class ArtCosto(BaseModel):
    sucursal = DummyForeignKey(Sucursal, db_column='Sucursal', to_field='sucursal', on_delete=models.PROTECT,
                               related_name='art_costos')
    empresa = models.CharField(db_column='Empresa', max_length=5)
    articulo = DummyForeignKey(Art, db_column='Articulo', to_field='articulo', on_delete=models.PROTECT,
                               related_name='art_costos')
    ultimo_costo = models.FloatField(db_column='UltimoCosto', null=True, blank=True)
    costo_promedio = models.FloatField(db_column='CostoPromedio', null=True, blank=True)
    costo_estandar = models.FloatField(db_column='CostoEstandar', null=True, blank=True)
    costo_reposicion = models.FloatField(db_column='CostoReposicion', null=True, blank=True)
    ultimo_costo_sin_gastos = models.FloatField(db_column='UltimoCostoSinGastos', null=True, blank=True)

    class Meta:
        managed = False
        db_table = 'ArtCosto'


class InterfacesPredefinidas(BaseModel):
    interfaz_id = models.IntegerField(unique=True)
    renglon_id = models.IntegerField(db_column='RenglonID', blank=True, null=True)
    interfaz = models.CharField(db_column='Interfaz', primary_key=True, max_length=100)
    marca = models.CharField(db_column='Marca', max_length=50)
    estatus = models.BooleanField(db_column='Estatus')
    reenvio_rango = models.BooleanField(db_column='ReenvioRango')
    abrir_mov = models.BooleanField(db_column='AbrirMov')
    reenvio_multiple = models.BooleanField(db_column='ReenvioMultiple')
    descripcion = models.CharField(db_column='Descripcion', max_length=100, blank=True, null=True)

    class Meta:
        managed = False
        db_table = 'CA_InterfacesPredefinidas'
        unique_together = (('interfaz_id', 'interfaz', 'marca'),)


class InterfacesPredefinidasDEmpresa(BaseModel):
    interfaz_id = DummyForeignKey(InterfacesPredefinidas, db_column='ID', to_field='interfaz_id',
                                  on_delete=models.PROTECT,
                                  related_name='interfaces')
    renglon_id = models.IntegerField(db_column='RenglonID', blank=True, null=True)
    clave = models.CharField(db_column='Clave', max_length=50)
    descripcion = models.CharField(db_column='Descripcion', max_length=100)
    valor_default = models.CharField(db_column='ValorDefault', max_length=255, blank=True, null=True)

    class Meta:
        managed = False
        db_table = 'CA_InterfacesPredefinidasDEmpresa'


class ParametrosSucursal(BaseModel):
    empresa = models.CharField(db_column='Empresa', max_length=5)
    sucursal = models.IntegerField(db_column='Sucursal')
    clave = models.CharField(db_column='Clave', max_length=30, primary_key=True)
    descripcion = models.CharField(db_column='Descripcion', max_length=200, blank=True, null=True)
    grupo = models.CharField(db_column='Grupo', max_length=20, blank=True, null=True)
    valor = models.CharField(db_column='Valor', max_length=100, blank=True, null=True)

    class Meta:
        managed = False
        db_table = 'CA_CatParametrosSucursal'
        unique_together = (('sucursal', 'clave'),)



from django.core.management import BaseCommand

from isapilib.api.models import ConnectionAPI


class Command(BaseCommand):
    help = 'Check connection to external database'

    def add_arguments(self, parser):
        parser.add_argument(
            "connection",
            help="Nominates a ConnectionAPI primary key to check connectivity",
            type=int,
        )

    def handle(self, *args, **options):
        try:
            connectionapi = ConnectionAPI.objects.get(pk=options["connection"])
            name = connectionapi.get_name()

            result, description = connectionapi.verify_connection()
            if result:
                self.stdout.write(self.style.SUCCESS(f'Connection to database {name} was successful.'))
            else:
                self.stdout.write(self.style.SUCCESS(f'Failed to connect to database {name}: {description}'))

        except ConnectionAPI.DoesNotExist:
            self.stdout.write(self.style.SUCCESS(f"Branch doesn't exist."))



from django.core.management.commands.inspectdb import Command as InspectDBCommand

from isapilib.api.models import ConnectionAPI


class Command(InspectDBCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            "connection",
            help="Nominates a ConnectionAPI primary key to introspect",
            type=int,
        )
        parser.add_argument(
            "table",
            nargs="*",
            type=str,
            help="Selects what tables or views should be introspected.",
        )
        parser.add_argument(
            "--include-partitions",
            action="store_true",
            help="Also output models for partition tables.",
        )
        parser.add_argument(
            "--include-views",
            action="store_true",
            help="Also output models for database views.",
        )

    def handle_inspection(self, options):
        connection_pk = options.pop("connection")
        connection = ConnectionAPI.objects.get(pk=connection_pk)
        options['database'] = connection.create_connection()
        return super().handle_inspection(options)

    def normalize_col_name(self, col_name, used_column_names, is_relation):
        new_name, field_params, field_notes = super().normalize_col_name(col_name, used_column_names, is_relation)
        return new_name, field_params, []



from functools import reduce

from django.conf import settings
from django.db import models
from rest_framework.fields import empty, Field, SkipField

from isapilib.core.exceptions import CreationError
from isapilib.external.utilities import get_param_sucursal, get_param_empresa
from isapilib.logging import logger


class CoalesceField(Field):
    def __init__(self, separator=' ', list_source=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.list_source = list_source
        self.separator = separator
        self.source_provided = False

    def bind(self, field_name, parent):
        self.source_provided = self.source is not None
        return super().bind(field_name, parent)

    def get_attribute(self, instance):
        values = []
        for source in self.list_source:
            for col in source.split('+'):
                if value := getattr(instance, col, None):
                    values.append(value)
                else:
                    values = []
                    break
            if 0 < len(values):
                break

        return self.separator.join(values) if len(values) > 0 else None

    def to_representation(self, value):
        return value

    def to_internal_value(self, data):
        return data


class JsonField(Field):
    def __init__(self, json_path=None, return_type=None, prefix=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.json_path = json_path
        self.return_type = return_type or str
        self.prefix = prefix

    def _get_json_value(self, dictionary):
        real_json_path = self.json_path or self.field_name

        try:
            return reduce(lambda d, key: d[key], real_json_path.split("."), dictionary)
        except (KeyError, TypeError):
            return empty

    def get_value(self, dictionary):
        value = self._get_json_value(dictionary)
        if self.prefix:
            value = f'{self.prefix}{'' if value is empty else value}'
        return value

    def to_internal_value(self, data):
        return self.return_type(data)

    def to_representation(self, value):
        return self.return_type(value)


class SerializerMethodField(Field):
    def __init__(self, method_name_get=None, method_name_set=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.method_name_get = method_name_get
        self.method_name_set = method_name_set
        self._value = empty

    def bind(self, field_name, parent):
        super().bind(field_name, parent)
        if self.method_name_get is None: self.method_name_get = f'get_{field_name}'
        if self.method_name_set is None: self.method_name_set = f'set_{field_name}'

    def get_attribute(self, instance):
        if method := getattr(self.parent, self.method_name_get, False):
            return method(instance)
        res = super().get_attribute(instance)
        return res

    def get_value(self, dictionary):
        if self._value is not empty: return self._value
        method = getattr(self.parent, self.method_name_set)
        self._value = method(dictionary)
        if self._value is empty or self._value is None:
            try:
                self._value = self.get_default()
            except SkipField:
                self._value = None
        return self._value

    def to_internal_value(self, data):
        return data

    def to_representation(self, value):
        if issubclass(type(value), models.Model):
            return value.pk
        else:
            return value


class GetOrField(JsonField):
    def __init__(self, model, model_field=None, related=True, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model = model
        self.model_field = model_field
        self.related = related

    def bind(self, field_name, parent):
        super().bind(field_name, parent)
        if self.model_field is None: self.model_field = field_name

    def to_internal_value(self, data):
        instance = self.model.objects.filter(pk=data).first()
        if instance is None: raise self.model.DoesNotExist
        return instance if self.related else instance.pk

    def to_representation(self, value):
        return getattr(value, 'pk', value)

    def get_value(self, dictionary):
        value = super().get_value(dictionary)
        if value is not empty and self.model.objects.filter(**{self.model_field: value}).exists():
            return self.model.objects.filter(**{self.model_field: value}).first().pk
        return empty


class GetOrDefaultField(GetOrField):
    def __init__(self, sucursal_param=None, empresa_param=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.sucursal_param = sucursal_param
        self.empresa_param = empresa_param

    def get_value(self, dictionary):
        value = super().get_value(dictionary)
        if value is not empty: return value

        model_name = self.model.__name__
        logger.debug('Value not found, attempting to get with parameter %s', model_name)

        pk = empty
        if self.sucursal_param:
            instance = self.parent.get_instance()
            pk = get_param_sucursal(instance.sucursal, self.sucursal_param)
            logger.debug('Instance of %s got (%s), sucursal %s, key %s',
                         model_name, pk, instance.sucursal, self.sucursal_param)

        if self.empresa_param:
            interfaz = getattr(settings, 'INTERFAZ_NAME', '')
            pk = get_param_empresa(interfaz, self.empresa_param)
            logger.debug('Instance of %s got (%s), interfaz %s, key %s',
                         model_name, pk, interfaz, self.empresa_param)

        return pk


class GetOrCreateField(GetOrField):
    def __init__(self, method_name=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.method_name = method_name

    def bind(self, field_name, parent):
        super().bind(field_name, parent)
        if self.method_name is None: self.method_name = f'create_{field_name}'

    def get_value(self, dictionary):
        value = super().get_value(dictionary)
        if value is not empty: return value

        model_name = self.model.__name__
        logger.debug('Value not found, attempting to create %s', model_name)

        try:
            method = getattr(self.parent, self.method_name)
            instance = method(self.parent.initial_data)
            if not isinstance(instance, self.model):
                raise ValueError(f'The function {self.method_name} did not return an instance of the model')
            instance.save()

            logger.debug('Instance of %s created successfully (%s)', model_name, instance.pk)

            return instance.pk
        except CreationError:
            return empty



from collections.abc import Iterable
from datetime import timezone, datetime, timedelta

from django.conf import settings
from django.contrib.auth.base_user import AbstractBaseUser, BaseUserManager
from django.db import connections
from django.db import models
from django.db.models.signals import post_save
from django.dispatch import receiver
from encrypted_model_fields.fields import EncryptedCharField

from isapilib.core.exceptions import PermissionDenied
from isapilib.core.utilities import is_test
from isapilib.logging import logger


class ConnectionAPI(models.Model):
    alias = models.CharField(max_length=40, null=True, blank=True, unique=True)
    user = models.CharField(max_length=40)
    password = EncryptedCharField(max_length=40)
    host = models.CharField(max_length=40)
    port = models.CharField(max_length=80, null=True, blank=True)
    nombre_db = models.CharField(max_length=80)
    empresa = models.CharField(max_length=80, null=True, blank=True)
    version = models.IntegerField(default=6000)
    allow_test = models.BooleanField(default=False)
    extra_data = models.JSONField(default=dict, null=True, blank=True)
    connection_timeout = models.IntegerField(default=0)
    connection_retries = models.IntegerField(default=5)
    connection_retry_backoff_time = models.IntegerField(default=5)
    query_timeout = models.IntegerField(default=0)

    def get_name(self):
        return f'external-{self.pk}'

    def get_database_configuration(self, refresh=True):
        if self.pk and refresh: self.refresh_from_db()

        options = settings.DATABASES['default'].get('OPTIONS', {})
        if not isinstance(options, dict): options = {}

        options.setdefault('connection_timeout', self.connection_timeout)
        options.setdefault('connection_retries', self.connection_retries)
        options.setdefault('connection_retry_backoff_time', self.connection_retry_backoff_time)
        options.setdefault('query_timeout', self.query_timeout)

        db_host = getattr(self, 'host', '')
        db_port = getattr(self, 'port', '')
        db_name = getattr(self, 'nombre_db', '')
        version = getattr(self, 'version', '')

        logger.debug('Generated configuration for external-%s, db: %s,%s %s %s',
                     self.pk, db_host, db_port or 1443, db_name, version)

        return {
            'ENGINE': 'mssql',
            'HOST': db_host,
            'PORT': db_port,
            'NAME': db_name,
            'USER': getattr(self, 'user', ''),
            'PASSWORD': getattr(self, 'password', ''),
            'INTELISIS_VERSION': version,
            'TIME_ZONE': None,
            'CONN_HEALTH_CHECKS': None,
            'CONN_MAX_AGE': None,
            'ATOMIC_REQUESTS': None,
            'AUTOCOMMIT': True,
            'OPTIONS': options
        }

    def create_connection(self):
        name = self.get_name()

        already_exists = name in connections.databases
        if not already_exists and is_test() and not self.allow_test:
            raise PermissionDenied(f'Test is not allowed at this connection {self.pk}')

        connections.databases[name] = self.get_database_configuration()

        status = 'updated' if already_exists else 'created'
        logger.info('Database connection for external-%s %s', self.pk, status)

        return name

    def verify_connection(self):
        name = 'test_connection'
        connections.databases[name] = self.get_database_configuration()
        data = ()

        try:
            connection = connections[name]
            cursor = connection.cursor()
            cursor.execute('SELECT 1')
            data = (True, None)
        except Exception as e:
            data = (False, str(e))
        finally:
            logger.debug('Connection test for external-%s %s', self.pk, 'connected' if data[0] else 'fail')

        return data

    class Meta:
        db_table = 'isapilib_connectionapi'
        unique_together = ('host', 'port', 'nombre_db')
        ordering = ['host', 'port', 'id']


class BranchAPI(models.Model):
    connection = models.ForeignKey(ConnectionAPI, on_delete=models.CASCADE, related_name='branches')
    sucursal = models.IntegerField()
    gwmbac = models.CharField(max_length=100, null=True, unique=True)

    def check_permissions(self, user):
        if user.is_anonymous:
            raise PermissionDenied(f'Anonymous user is not allowed at this connection {self.pk}')
        return self in user.permissions.all()

    class Meta:
        db_table = 'isapilib_branchapi'
        ordering = ['connection', 'id']


class UserProfileManager(BaseUserManager):
    def create_user(self, **kwargs):
        user = self.model.objects.create(**kwargs)
        user.set_password(kwargs['password'])
        user.save()
        return user

    def create_superuser(self, **kwargs):
        kwargs.setdefault('is_superuser', True)
        return self.create_user(**kwargs)


class UserAPI(AbstractBaseUser):
    username = models.CharField(max_length=30, unique=True)
    is_superuser = models.BooleanField(default=False)
    branch = models.ForeignKey(BranchAPI, on_delete=models.CASCADE, null=True, blank=True, related_name='users')
    permissions = models.ManyToManyField(BranchAPI, related_name="permissions")

    USERNAME_FIELD = 'username'

    objects = UserProfileManager()

    class Meta:
        db_table = 'isapilib_userapi'
        ordering = ['username', 'id']


class ApiLogs(models.Model):
    user = models.ForeignKey(UserAPI, on_delete=models.CASCADE, null=True, blank=True, related_name='logs')
    tipo = models.CharField(max_length=120)
    header = models.TextField()
    request = models.TextField()
    response = models.TextField()
    status = models.IntegerField(null=True)
    url = models.TextField()
    interfaz = models.CharField(max_length=120)
    fecharegistro = models.DateTimeField()
    response_time = models.IntegerField(default=0)

    class Meta:
        db_table = 'api_logs'
        ordering = ['-fecharegistro', 'id']


@receiver(post_save, sender=ApiLogs)
def delete_old_logs(sender, instance, **kwargs):
    interfaz = getattr(settings, 'INTERFAZ_NAME', '')
    max_number_logs = getattr(settings, 'MAX_NUMBER_LOGS', timedelta(days=30))

    logs_to_delete = None

    if isinstance(max_number_logs, int):
        logs_to_delete = ApiLogs.objects.filter(interfaz=interfaz).order_by('-fecharegistro')[max_number_logs:]

    if isinstance(max_number_logs, timedelta):
        date_to_delete = datetime.now(tz=timezone.utc) - max_number_logs
        logs_to_delete = ApiLogs.objects.filter(interfaz=interfaz, fecharegistro__lt=date_to_delete)

    if not isinstance(logs_to_delete, Iterable):
        return

    deleted = 0
    for log in logs_to_delete:
        count, _ = log.delete()
        deleted += count

    if 0 < deleted: logger.info('%s logs cleanup completed, Removed %s entries', interfaz, deleted)



from django.core.management.base import BaseCommand

from isapilib.api.models import UserAPI, BranchAPI


class Command(BaseCommand):
    help = 'Allow user to connect to database'

    def add_arguments(self, parser):
        parser.add_argument("username", help="Username of user", type=str)
        parser.add_argument("branch", help="Branch api", type=str)

    def handle(self, *args, **options):
        try:
            user = UserAPI.objects.get(username=options['username'])
            branch = BranchAPI.objects.get(pk=options["branch"])

            user.permissions.add(branch)
            self.stdout.write(self.style.SUCCESS(f'{user.username} now can connect to external-{branch.pk}'))
        except UserAPI.DoesNotExist:
            self.stdout.write(self.style.SUCCESS(f"User doesn't exist."))

        except BranchAPI.DoesNotExist:
            self.stdout.write(self.style.SUCCESS(f"Database doesn't exist."))



from django.conf import settings
from django.db import models
from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from isapilib import validators
from isapilib.core.utilities import SerializerData, get_sucursal_from_request
from isapilib.external.utilities import get_almacen, get_sucursal, get_uen
from isapilib.models import Venta, Empresa, Vin, VentaD, Art, CaVenta


def nested_json(attrs: dict, *args, **kwargs):
    serializer = type('NestedJson', (serializers.Serializer,), attrs)
    return serializer(*args, **kwargs)


class CreateSerializer(serializers.Serializer):
    required_fields = []
    priority_fields = []

    _meta = None

    def __init__(self, *args, **kwargs):
        model = getattr(getattr(self, 'Meta', None), 'model', None)
        meta_class = {field: self.get_mapping(model, field) for field in self.required_fields}
        meta_class['Meta'] = getattr(self, 'Meta')
        setattr(self, 'Meta', None)

        super().__init__(*args, **kwargs)
        self.model_serializer_class = type('ModelSerializer', (serializers.ModelSerializer,), meta_class)
        self.serializer = None
        self.base_data = None

    def get_mapping(self, model, field_name):
        field = getattr(getattr(model, field_name), 'field')
        return serializers.ModelSerializer.serializer_field_mapping[type(field)]

    @property
    def meta(self):
        if not hasattr(self, '_meta'):
            self._meta = getattr(self, "Meta", None)

        return self._meta

    @staticmethod
    def normalize_data(data):
        normalized_data = {}
        for name, value in data.items():
            if isinstance(value, models.Model):
                normalized_data[name] = value.pk
            else:
                normalized_data[name] = value

        return normalized_data

    def transform_data(self, attrs):
        self.base_data = SerializerData(attrs)
        serializer = self.model_serializer_class()
        fields = list(dict.fromkeys(self.priority_fields + list(serializer.fields.keys())))

        data = {}
        for field in fields:
            method = getattr(self, "get_%s" % field, None)
            if not method: continue
            data[field] = method(data)

        return data

    def validate(self, attrs):
        transformed_attrs = self.transform_data(attrs)

        try:
            transformed_attrs = self.normalize_data(transformed_attrs)
            self.serializer = self.model_serializer_class(data=transformed_attrs)
            self.serializer.is_valid(raise_exception=True)
        except ValidationError as e:
            raise serializers.ValidationError({
                "detail": "Error while transforming input data to model format",
                "model_errors": e.detail
            })

        return transformed_attrs

    def create(self, validated_data):
        return self.serializer.save()

    def to_representation(self, instance):
        return self.serializer.data


class CreateVentaSerializer(CreateSerializer):
    priority_fields = ['mov', 'sucursal']
    afectar = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        attrs = {'Meta': type('Meta', (), {'model': VentaD, 'fields': '__all__'})}
        self.detail_serializer_class = type('DetailSerializer', (serializers.ModelSerializer,), attrs)

        attrs = {'Meta': type('Meta', (), {'model': CaVenta, 'fields': '__all__'})}
        self.ca_venta_serializer_class = type('CAVentaSerializer', (serializers.ModelSerializer,), attrs)

    def get_empresa(self, _):
        return Empresa.objects.first()

    def get_sucursal(self, data):
        request = self.context['request']
        return get_sucursal(mov=data['mov'], sucursal=get_sucursal_from_request(request))

    def get_sucursal_origen(self, data):
        return data['sucursal']

    def get_almacen(self, data):
        return get_almacen(mov=data['mov'], sucursal=data['sucursal'])

    def get_uen(self, data):
        return get_uen(mov=data['mov'], sucursal=data['sucursal'])

    def get_referencia(self, data):
        return getattr(settings, 'INTERFAZ_NAME', '')

    def transform_data(self, validated_data):
        transformed_data = super().transform_data(validated_data)

        vin = transformed_data.get('servicio_serie')
        if isinstance(vin, Vin):
            transformed_data['servicio_kms'] = vin.km
            transformed_data['servicio_modelo'] = vin.modelo
            transformed_data['servicio_placas'] = vin.placas
            transformed_data['servicio_articulo'] = vin.articulo
            transformed_data['servicio_descripcion'] = vin.color_exterior
            transformed_data['servicio_identificador'] = vin.color_exterior

        self.context['transformed_data'] = transformed_data
        return transformed_data

    def create(self, validated_data):
        instance: Venta = super().create(validated_data)
        transformed_data = self.context['transformed_data']

        if get_details := getattr(self, 'get_details', None):
            for detail in get_details(transformed_data):
                try:
                    self.create_detail(instance, detail)
                except ValidationError as e:
                    raise ValidationError({'details': e.detail})

        if get_ca_venta := getattr(self, 'get_ca_venta', None):
            ca_venta_data = get_ca_venta(transformed_data)
            ca_venta_data['venta'] = instance.pk

            try:
                serializer = self.ca_venta_serializer_class(data=ca_venta_data)
                serializer.is_valid(raise_exception=True)
                serializer.save()
            except ValidationError as e:
                raise ValidationError({'ca_venta': e.detail})

        if self.afectar: instance.afectar(accion=self.afectar)
        return instance

    def get_detail_data(self, instance: Venta, data) -> dict:
        count = VentaD.objects.filter(venta=instance).count() + 1

        articulo: Art = data['articulo']
        cantidad = data.get('cantidad', 1)

        return {
            'renglon_id': count,
            'renglon': 2048 * count,

            'venta': instance,
            'uen': instance.uen,
            'agente': instance.agente,
            'almacen': instance.almacen,
            'sucursal': instance.sucursal,
            'sucursal_origen': instance.sucursal,
            'hora_requerida': instance.hora_requerida,

            'articulo': articulo,
            'impuesto1': articulo.impuesto1,
            'descripcion_extra': articulo.descripcion1,
            'comentarios': articulo.descripcion1,
            'precio': articulo.precio_lista or 0,
            'cantidad': cantidad,

            'cc_tiempo_tab': 1,
            'renglon_tipo': 'N',
            'renglon_sub': 0,
            'ut': 1,
        }

    def create_detail(self, instance: Venta, data):
        base_data = self.get_detail_data(instance, data)
        base_data.update(data)

        detail_data = self.normalize_data(base_data)
        serializer = self.detail_serializer_class(data=detail_data)
        serializer.is_valid(raise_exception=True)
        serializer.save()

    class Meta:
        model = Venta
        fields = '__all__'


class CreateAppointmentSerializer(CreateVentaSerializer):
    required_fields = ['fecha_requerida', 'fecha_emision', 'hora_requerida', 'hora_recepcion']
    priority_fields = CreateVentaSerializer.priority_fields + ['fecha_requerida', 'hora_requerida']

    def get_mov(self, _):
        return 'Cita Servicio'

    def get_fecha_emision(self, data):
        return data['fecha_requerida']

    def get_hora_recepcion(self, data):
        return data['hora_requerida']

    class Meta(CreateVentaSerializer.Meta):
        validators = [
            validators.DateAfterTodayValidator(),
        ]



import io
import sys
import time
import traceback

from django.conf import settings
from django.contrib.auth.models import AnonymousUser
from django.http import HttpResponse
from rest_framework.renderers import JSONRenderer
from rest_framework.response import Response
from rest_framework.settings import api_settings
from rest_framework.views import APIView

from isapilib.auth.permissions import IsapilibPermission
from isapilib.core.get_functions import get_insert_log_function, get_safe_method_function
from isapilib.logging import logger

insert_log = get_insert_log_function()
safe_method = get_safe_method_function()


class LoggerMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response
        self.get_exception_response = safe_method(self.launch_exception)

        self.interfaz = getattr(settings, 'INTERFAZ_NAME', '')

        self.authentication_classes = api_settings.DEFAULT_AUTHENTICATION_CLASSES
        self.authenticators = [auth() for auth in self.authentication_classes]

    @staticmethod
    def launch_exception(exception):
        raise exception

    @staticmethod
    def get_log_type(view_func) -> str:
        view_class = getattr(view_func, 'view_class', None)
        name = view_class.__name__ if view_class else view_func.__name__

        for suffix in ['View', 'ViewSet']:
            if name.endswith(suffix):
                name = name[:-len(suffix)]

        return name

    def process_view(self, base_request, view_func, view_args, view_kwargs):
        base_request.user = getattr(base_request, 'user', AnonymousUser())
        base_request.log_type = self.get_log_type(view_func)

        view_class = getattr(view_func, 'view_class', getattr(view_func, 'cls', None))
        permission_classes: tuple = getattr(view_class, 'permission_classes', ())
        base_request.is_drf = view_class is not None and issubclass(view_class, APIView)
        base_request.log_transaction = IsapilibPermission in permission_classes

    def process_exception(self, _, exception):
        tb = traceback.format_exc().strip().replace("\n", " | ")
        logger.error("Exception detected exception=%s: %s traceback=%s", type(exception).__name__, str(exception), tb)

        response = self.get_exception_response(exception)

        if not isinstance(response, Response):
            return response

        response.accepted_renderer = JSONRenderer()
        response.accepted_media_type = "application/json"
        response.renderer_context = {}

        response.render()
        return HttpResponse(
            content=response.content,
            status=response.status_code,
            content_type=response.content_type,
        )

    def __call__(self, base_request):
        IsapilibPermission.reset_current_request()

        start = time.time()
        response = self.get_response(base_request)
        end = time.time()

        is_drf = getattr(base_request, 'is_drf', False)
        request = IsapilibPermission.get_current_request(raise_exception=False)

        if not is_drf or request is None or not base_request.log_transaction:
            return response

        user = request.user

        if user.is_authenticated:
            log_type = f'{request.log_type} {request.method}'
            response_time = (end - start) * 1000

            logger.debug('Logged %s %s %sms', self.interfaz, log_type, response_time)

            insert_log(
                request=request,
                response=response,
                interfaz=self.interfaz,
                tipo=log_type,
                time=response_time,
            )

        return response


class ChunkedMiddleware:
    def __init__(self, get_response):
        development = any(arg for arg in sys.argv if "manage.py" in arg)
        support = any(s in arg for arg in sys.argv for s in ("uvicorn", "hypercorn", "daphne"))

        if not support and not development:
            raise Exception('Chunked request not supported. Use a server that handles Transfer-Encoding.')

        self.get_response = get_response

    def __call__(self, request):
        has_transfer_encoding = 'Transfer-Encoding' in request.headers
        transfer_encoding = request.headers.get('Transfer-Encoding')

        if has_transfer_encoding and transfer_encoding == 'chunked':
            logger.info('Chunked request received chunked transfer encoding')

            content_length = request.META.get('CONTENT_LENGTH')
            if content_length in (None, '0'):
                content_length = str(len(request.body))
                request.META['CONTENT_LENGTH'] = content_length
                request._stream = io.BytesIO(request.body)
                logger.debug('Transformed chunked request to standard encoding, set CONTENT_LENGTH=%s', content_length)

        response = self.get_response(request)
        return response



import sys

from django.conf import settings


def get_default_user():
    return getattr(settings, 'INTELISIS_USER', 'SOPDESA')


def get_dealer_field():
    return getattr(settings, 'DEALER_FIELD', 'dealerID')


def to_bool(s):
    if isinstance(s, bool):
        return s
    if isinstance(s, str):
        s = s.lower()
        if s in ['true', '1', 'yes']:
            return True
        elif s in ['false', '0', 'no']:
            return False
        else:
            return False
    return False


def get_sucursal_from_request(request):
    try:
        return request.user.branch.sucursal
    except Exception as e:
        return 0


def is_test():
    return sys.argv[1] == 'test'


class SerializerData(dict):
    def get(self, string_keys: str, default=None):
        value = dict(self)
        for key in string_keys.split('.'):
            value = value.get(key, default) if isinstance(value, dict) else default
        return value



# Generated by Django 5.2.5 on 2025-10-03 20:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('isapilib', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='apilogs',
            name='fecharegistro',
            field=models.DateTimeField(),
        ),
    ]



from django.conf import settings
from django.utils.module_loading import import_string


def get_request_content_function():
    try:
        get_request_content = import_string(settings.GET_REQUEST_CONTENT_FUNCTION)
    except Exception:
        from isapilib.logger.functions import get_request_content

    return get_request_content


def get_response_content_function():
    try:
        get_response_content = import_string(settings.GET_RESPONSE_CONTENT_FUNCTION)
    except Exception:
        from isapilib.logger.functions import get_response_content

    return get_response_content


def get_insert_log_function():
    try:
        insert_log = import_string(settings.INSERT_LOG_FUNCTION)
    except Exception:
        from isapilib.logger.create_log import insert_log

    return insert_log


def get_safe_method_function():
    try:
        safe_method = import_string(settings.SAFE_METHOD_FUNCTION)
    except Exception:
        from isapilib.core.decorators import safe_method

    return safe_method



from __future__ import annotations

import random
from datetime import datetime, timedelta
from typing import Union, TYPE_CHECKING, List

from rest_framework import status

from isapilib.core import exceptions
from isapilib.core.utilities import get_default_user
from isapilib.external.utilities import execute_sp, get_utc_offset
from isapilib.mov.traspaso import TraspasoItems

if TYPE_CHECKING:
    from isapilib.models import Venta, VentaD


class VentaMixin:
    class ACTIONS:
        AFECTAR = 'AFECTAR'
        CONSECUTIVO = 'CONSECUTIVO'
        GENERAR = 'GENERAR'
        CANCELAR = 'CANCELAR'

    def get_origen(self: Venta) -> Union[Venta, None]:
        from isapilib.models import Venta
        if self.origen is None or self.origen_id is None: return None
        return Venta.objects.filter(mov=self.origen, mov_id=self.origen_id).order_by('-fecha_emision').first()

    def get_destinations(self: Venta) -> Union[Venta, None]:
        from isapilib.models import Venta
        if self.mov is None or self.mov_id is None: return None
        return Venta.objects.filter(origen=self.mov, origen_id=self.mov_id).order_by('-fecha_emision').all()

    def create_parts_order_from_service(self: Venta, usuario=None) -> Venta:
        from isapilib.models import Venta
        usuario = usuario or get_default_user()

        if self.mov != 'Servicio': raise ValueError('La instancia de venta no es un servicio')

        if self.estatus != 'PENDIENTE':
            msg = 'El servicio no esta afectado' if self.estatus == 'SINAFECTAR' else 'El servicio ya esta concluido'
            raise exceptions.RequestError(msg)

        estimated_pk = execute_sp('xpCA_xpMovCopiarSoloEncabezado', {
            'Sucursal': self.sucursal.pk,
            'Modulo': 'VTAS',
            'ID': self.id,
            'Empresa': self.empresa.pk,
            'Mov': 'Pedido Servicio',
            'MovID': self.mov_id,
            'Usuario': usuario,
            'TipoCambio': self.tipo_cambio,
            'Moneda': self.moneda,
            'Almacen': self.almacen.pk,
            'AlmacenDestino': self.almacen.pk,
        })[0][0]

        estimated = Venta.objects.get(id=estimated_pk)
        estimated.origen = self.mov
        estimated.origen_id = self.mov_id
        estimated.save()
        return estimated

    def create_service_from_appointment(self: Venta, action=None, usuario=None) -> Venta:
        from isapilib.models import Venta
        usuario = usuario or get_default_user()

        if self.mov != 'Cita Servicio': raise ValueError('La instancia de venta no es una cita servicio')
        query_order = Venta.objects.filter(origen=self.mov, origen_id=self.mov_id, mov='Servicio').order_by(
            '-fecha_emision')

        order = query_order.first()

        if not self.details.exists():
            raise exceptions.RequestError('La cita no tiene detalles')

        if self.estatus == 'SINAFECTAR':
            self.afectar()

        if order is None:
            code, description = self.afectar(accion=self.ACTIONS.GENERAR, generar='Servicio', usuario=usuario)
            if code != 80030:
                raise exceptions.CoreException(code, description, self)  # ehre
            order = query_order.first()

        if action: order.afectar(accion=action)
        return order

    def traspasar_items(self, details: List[VentaD]):
        if len(details) == 0: return

        order = self.get_origen()
        estacion = random.randint(10000, 99999)

        traspaso = TraspasoItems(order, estacion)

        for detail in details:
            traspaso.add_item(detail)

        traspaso.execute()

    def accept_parts_order(self: Venta, items: tuple[str, bool]) -> bool:
        from isapilib.models import VentaD

        if self.mov != 'Pedido Servicio': raise ValueError('La instancia de venta no es una pedido servicio')
        items_pk = []
        errors = []

        for art_pk, is_accept in items:
            try:
                self.details.get(articulo=art_pk)
                if is_accept: items_pk.append(art_pk)
            except VentaD.DoesNotExist:
                errors.append((art_pk, 'El articulo no existe en el pedido servicio'))

        if 0 < len(errors):
            raise exceptions.RequestError(errors)

        items = list(self.details.filter(articulo__in=items_pk))
        self.traspasar_items(items)

    def cambiar_situacion(self: Venta, situacion, modulo='VTAS', fecha=None, usuario=None, situacion_usuario=None,
                          situacion_nota=None):
        usuario = usuario or get_default_user()

        if fecha is None:
            fecha = datetime.now()
            fecha += timedelta(minutes=get_utc_offset())

        if isinstance(fecha, datetime):
            fecha = fecha.strftime('%Y-%m-%d %H:%M:%S.') + f'{int(fecha.microsecond / 1000):03d}'

        params = [modulo, self.pk, situacion, fecha, usuario, situacion_usuario, situacion_nota]
        execute_sp('spCambiarSituacion', params)
        self.refresh_from_db()
        if self.situacion != situacion: raise Exception(f'Error al cambiar la situacion ({self.pk})')

    def afectar(self: Venta, *, accion=None, generar=None, usuario=None, s=0, modulo=None, estacion=None, base=None):
        usuario = usuario or get_default_user()

        accion = accion or self.ACTIONS.AFECTAR
        code, description = execute_sp('spAfectar', {
            'Modulo': modulo or 'VTAS',
            'ID': self.id,
            'Accion': accion,
            'Base': base or 'Todo',
            'Usuario': usuario,
            'GenerarMov': generar,
            'EnSilencio': s,
            'Estacion': estacion or 1000,
        })[0]

        self.refresh_from_db()

        if (
                (accion == self.ACTIONS.AFECTAR and self.estatus == 'SINAFECTAR') or
                (accion == self.ACTIONS.CONSECUTIVO and self.mov_id is None)
        ):
            raise exceptions.AffectationError(code, description, self.pk, status=status.HTTP_409_CONFLICT)

        return code, description

    def cancelar(self):
        return self.afectar(accion=self.ACTIONS.CANCELAR)



import random
import string
import unittest
from datetime import timedelta, datetime
from typing import Union

from django.conf import settings
from django.contrib.auth import get_user_model
from django.db.models import QuerySet
from django.test import RequestFactory
from django.utils import timezone
from rest_framework import status
from rest_framework.test import APIClient

from isapilib.api.models import ApiLogs
from isapilib.auth.permissions import IsapilibPermission
from isapilib.core.utilities import get_dealer_field


def set_headers(dealer, fn):
    def wrapper(*args, **kwargs):
        dealer_field = get_dealer_field()
        headers = kwargs.get('headers', {})
        headers[dealer_field] = dealer
        return fn(*args, **kwargs, headers=headers)

    return wrapper


class IsapiTestCase(unittest.TestCase):
    interfaz = getattr(settings, 'INTERFAZ_NAME', None)
    auth_data: Union[dict, None] = None
    context = {}

    dealer = None

    class AUTHMETHODS:
        JWT = 'jwt'
        OAUTH2 = 'oauth2'
        AUTHTOKEN = 'authtoken'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.client = APIClient()
        self.factory = RequestFactory()

        if self.dealer is not None:
            dealer_field = get_dealer_field()
            setattr(self.client, 'put', set_headers(self.dealer, self.client.put))
            setattr(self.client, 'get', set_headers(self.dealer, self.client.get))
            setattr(self.client, 'post', set_headers(self.dealer, self.client.post))
            setattr(self.client, 'delete', set_headers(self.dealer, self.client.delete))

            setattr(settings, dealer_field, self.dealer)

        if isinstance(self.auth_data, dict):
            user = get_user_model()
            try:
                self.user = user.objects.get(username=self.auth_data['user'])
            except user.DoesNotExist:
                raise Exception(f'El usuario no existe {self.auth_data['user']}')
            self.authenticate()

            factory = RequestFactory()
            request = factory.get('')
            request.user = self.user
            IsapilibPermission.set_current_request(request)

    @property
    def logs(self) -> QuerySet[ApiLogs]:
        if not hasattr(self, '_production_logs'):
            self._production_logs = ApiLogs.objects.filter(interfaz=self.interfaz)
        return self._production_logs

    @staticmethod
    def log(msg):
        now = datetime.now().strftime('%I:%M %p').lower()
        print(msg, now)

    @staticmethod
    def generate_token(length=40):
        chars = string.ascii_letters + string.digits
        return ''.join(random.choice(chars) for _ in range(length))

    def authenticate(self):
        auth_method = self.auth_data.get('type', self.AUTHMETHODS.JWT)
        getattr(self, f'auth_{auth_method}')()

    def auth_oauth2(self):
        from oauth2_provider.models import AccessToken, Application

        access_token = AccessToken.objects.filter(user_id=self.user.pk, expires__gt=timezone.now()).first()

        if access_token is None:
            self.application = Application.objects.first()
            access_token = AccessToken.objects.create(
                user=self.user,
                application=self.application,
                token=self.generate_token(),
                expires=timezone.now() + timedelta(days=1),
            )
        else:
            self.application = access_token.application

        self.token = access_token.token
        prefix = self.auth_data.get('token_prefix', 'Bearer')
        self.client.credentials(HTTP_AUTHORIZATION=f'{prefix} ' + self.token)

    def auth_jwt(self):
        from rest_framework_simplejwt.tokens import RefreshToken

        refresh = RefreshToken.for_user(self.user)
        self.token = str(refresh.access_token)
        prefix = self.auth_data.get('token_prefix', 'Bearer')
        self.client.credentials(HTTP_AUTHORIZATION=f'{prefix} ' + self.token)

    def auth_authtoken(self):
        from rest_framework.authtoken.models import Token

        self.token, _ = Token.objects.get_or_create(user=self.user)
        prefix = self.auth_data.get('token_prefix', 'Token')
        self.client.credentials(HTTP_AUTHORIZATION=f'{prefix} ' + self.token.key)

    def assertResponse(self, request, response, status_code=status.HTTP_200_OK):
        self.assertEqual(
            status_code,
            response.status_code,
            f'\nURL: {response.request['PATH_INFO']}\nREQUEST: {request}\nRESPONSE: {response.content}'
        )



import json


def get_request_content(request):
    try:
        if isinstance(request.data, str):
            return request.data

        if isinstance(request.data, (dict, list)):
            return json.dumps(request.data, ensure_ascii=False)
    except Exception as e:
        return f'Warning: Failed to get request content, ({type(e)}) {e}'


def get_response_content(request, response):
    try:
        if hasattr(response, 'data'):
            try:
                return response.rendered_content.decode("utf-8", errors="replace")
            except AssertionError:
                response.accepted_renderer = request.accepted_renderer
                response.accepted_media_type = request.accepted_media_type
                response.renderer_context = {'view': None, 'args': (), 'kwargs': {}, 'request': request}

                return response.rendered_content.decode("utf-8", errors="replace")

        return response.content.decode('utf-8')
    except Exception as e:
        return f'Warning: Failed to get response content, ({type(e)}) {e}'



import functools
import traceback

from django.conf import settings
from rest_framework import status
from rest_framework.exceptions import APIException
from rest_framework.response import Response

from isapilib.core.exceptions import RequestError
from isapilib.core.utilities import is_test


def safe_method(view_func):
    @functools.wraps(view_func)
    def wrapped_view(*args, **kwargs):
        try:
            return view_func(*args, **kwargs)
        except APIException as e:
            raise e

        except RequestError as e:
            if is_test() or settings.DEBUG: traceback.print_exc()
            return Response({
                'message': str(e)
            }, content_type='application/json', status=e.status)

        except Exception as e:
            if is_test() or settings.DEBUG: traceback.print_exc()
            return Response({
                'type': str(type(e)),
                'message': str(e)
            }, content_type='application/json', status=status.HTTP_500_INTERNAL_SERVER_ERROR)

    return wrapped_view



from __future__ import annotations

from typing import TYPE_CHECKING

from django.db.models.aggregates import Sum

if TYPE_CHECKING:
    from isapilib.models import Vin


class VinMixin:
    def get_costo(self: Vin) -> int:
        from isapilib.models import Compra

        if isinstance(self.compra, Compra):
            return (self.compra.importe or 0) - (self.compra.impuestos or 0)
        return 0

    def get_precio(self: Vin) -> int:
        accessories = self.accessories.filter(tipo__afecta_precio=1)
        return accessories.aggregate(total=Sum('precio_contado'))['total'] or 0



from rest_framework.status import HTTP_400_BAD_REQUEST


class IsapiException(Exception):
    pass


class PermissionDenied(IsapiException):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return f'{self.message}'


class CreationError(IsapiException):
    pass


class RequestError(IsapiException):
    def __init__(self, message, status=None):
        self.status = status or HTTP_400_BAD_REQUEST
        self.message = message

    def __str__(self):
        return f'{self.message}'


class AffectationError(RequestError):
    def __init__(self, code, description, instance_pk=None, status=None):
        self.status = status or HTTP_400_BAD_REQUEST
        self.description = description
        self.instance_pk = instance_pk
        self.code = code

    def __str__(self):
        from isapilib.models import MensajeLista
        message = MensajeLista.objects.filter(mensaje=self.code).first()
        message = f'{message.descripcion}, {self.description}' if message else self.description
        return f'{self.code}: {message} ({self.instance_pk})'


class ProcessError(IsapiException):
    def __init__(self, instance):
        self.instance = instance


class TraspasarArticuloError(ProcessError):
    def __init__(self, instance, *args, **kwargs):
        super().__init__(instance)
        self.kwargs = kwargs
        self.args = args



from datetime import datetime, timedelta, UTC

from isapilib.core.exceptions import RequestError
from isapilib.external.utilities import get_utc_offset
from isapilib.models import Venta


class DateAfterTodayValidator:
    requires_context = False

    def __call__(self, value):
        hora_string = value['hora_requerida']
        hora_requerida = datetime.strptime(hora_string, '%H:%M').time()
        fecha_requerida = value['fecha_requerida']
        fecha_requerida = fecha_requerida.combine(fecha_requerida, hora_requerida)
        fecha_requerida = fecha_requerida.replace(tzinfo=UTC)
        current_date = datetime.now(UTC) + timedelta(minutes=get_utc_offset())
        try:
            if fecha_requerida < current_date:
                raise RequestError('The date must be after the current date')
        except TypeError as e:
            raise Exception(f'[{fecha_requerida.tzinfo}, {current_date.tzinfo}] - {e}')

        return value


class NoDuplicateAppointmentsValidator:
    requires_context = False

    def __call__(self, value):
        filters = {
            'mov': 'Cita Servicio',
            'agente': value['agente'],
            'fecha_requerida__date': value['fecha_requerida'].date(),
            'hora_recepcion': value['hora_recepcion'],
            'estatus': 'CONFIRMAR'
        }

        if Venta.objects.filter(**filters).exists():
            raise RequestError('This appointment has already been scheduled')



import threading

from rest_framework.permissions import BasePermission

from isapilib.core.exceptions import IsapiException
from isapilib.logging import logger

_thread_local = threading.local()


class IsapilibPermission(BasePermission):
    def has_permission(self, request, view):
        self.set_current_request(request)
        return True

    @staticmethod
    def get_current_request(raise_exception=True):
        request = getattr(_thread_local, 'request', None)
        if request is None and raise_exception:
            raise IsapiException(
                'Request is not available in thread-local context. '
                'Make sure IsapilibPermission is set either in DEFAULT_PERMISSION_CLASSES '
                'or included in the view\'s permission_classes.'
            )
        return request

    @staticmethod
    def set_current_request(request):
        logger.debug('Current request set in thread-local storage %s', id(request))
        _thread_local.request = request

    @staticmethod
    def reset_current_request():
        logger.debug('Thread-local request cleared')
        _thread_local.request = None



import sys

from django.core import exceptions
from django.core.management.base import BaseCommand
from django.db import models
from django.utils.text import capfirst


class NotRunningInTTYException(Exception):
    pass


class BaseCreateModelCommand(BaseCommand):
    Model = None

    def add_arguments(self, parser):
        parser.add_argument(
            "--all_fields",
            help="Specifies whether the record should be created using only the minimum required fields or all available fields",
            default=False,
        )

    def is_required(self, field):
        return not field.blank and field.default is models.NOT_PROVIDED

    def is_foreign_key(self, field):
        return field.is_relation and field.many_to_one

    def _get_input_message(self, field, default=None):
        return "%s%s%s: " % (
            capfirst(field.verbose_name),
            " (leave blank to use '%s')" % default if default else "",
            (
                " (%s.%s)"
                % (
                    field.remote_field.model._meta.object_name,
                    (
                        field.m2m_target_field_name()
                        if field.many_to_many
                        else field.remote_field.field_name
                    ),
                )
                if field.remote_field
                else ""
            ),
        )

    def get_input_data(self, field, message, default=None):
        raw_value = input(message)
        if default and raw_value == "":
            raw_value = default
        try:
            val = field.clean(raw_value, None)
        except exceptions.ValidationError as e:
            self.stderr.write("Error: %s" % "; ".join(e.messages))
            val = None

        return val

    def handle(self, *args, **options):
        if hasattr(sys.stdin, "isatty") and not sys.stdin.isatty():
            raise NotRunningInTTYException

        all_fields = options["all_fields"]

        data = {}

        fields = self.Model._meta.fields
        if not all_fields:
            fields = [i for i in fields if self.is_required(i)]

        for field in fields:
            if field.primary_key:
                continue

            field_name = field.name
            data[field_name] = None
            while data[field_name] is None:
                message = self._get_input_message(field)
                input_value = self.get_input_data(field, message)
                data[field_name] = input_value

                if field.many_to_many and input_value:
                    if not input_value.strip():
                        data[field_name] = None
                        self.stderr.write("Error: This field cannot be blank.")
                        continue
                    data[field_name] = [pk.strip() for pk in input_value.split(",")]

            if self.is_foreign_key(field):
                data[f"{field_name}_id"] = data.pop(field_name)

        instance = self.Model.objects.using('default').create(**data)

        self.stdout.write(self.style.SUCCESS(f"{self.Model.__name__} ({instance.pk}) created successfully."))



from typing import Any

from django.contrib.auth.models import update_last_login
from rest_framework_simplejwt import serializers as auth_serializers
from rest_framework_simplejwt.settings import api_settings
from rest_framework_simplejwt.tokens import RefreshToken


class TokenObtainPairSerializer(auth_serializers.TokenObtainSerializer):
    token_class = RefreshToken

    def validate(self, attrs: dict[str, Any]) -> dict[str, str]:
        data = super().validate(attrs)

        refresh = self.get_token(self.user)

        data["refresh_token"] = str(refresh)
        data["access_token"] = str(refresh.access_token)
        data["expires_in"] = int(refresh.access_token.lifetime.total_seconds())

        if api_settings.UPDATE_LAST_LOGIN:
            update_last_login(None, self.user)

        return data



import toml
from setuptools import setup, find_packages

config = toml.load('pyproject.toml')

setup(
    name=config['project']['name'],
    version=config['project']['version'],
    packages=find_packages(),
    include_package_data=True,
    package_data={},
    install_requires=[
        'django>=4.2',
        'djangorestframework>=3.0.0',
        'djangorestframework-simplejwt>=5.0.0',
        'djangorestframework-simplejwt>=5.0.0',
        'mssql-django>=1.5',
    ],
)



# Generated by Django 5.2.8 on 2025-12-03 14:37

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('isapilib', '0002_alter_apilogs_fecharegistro'),
    ]

    operations = [
        migrations.CreateModel(
            name='CuentasPorCobrar',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('mov', models.CharField(db_column='Mov', max_length=20)),
                ('mov_id', models.CharField(blank=True, db_column='MovID', max_length=20, null=True)),
            ],
            options={
                'db_table': 'Cxc',
                'managed': False,
            },
        ),
        migrations.AlterModelOptions(
            name='apilogs',
            options={'ordering': ['-fecharegistro', 'id']},
        ),
        migrations.AlterField(
            model_name='branchapi',
            name='gwmbac',
            field=models.CharField(max_length=100, null=True, unique=True),
        ),
    ]



# Generated by Django 5.2.5 on 2025-12-04 20:33

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('isapilib', '0003_cuentasporcobrar_alter_apilogs_options_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='connectionapi',
            name='connection_retries',
            field=models.IntegerField(default=5),
        ),
        migrations.AddField(
            model_name='connectionapi',
            name='connection_retry_backoff_time',
            field=models.IntegerField(default=5),
        ),
        migrations.AddField(
            model_name='connectionapi',
            name='connection_timeout',
            field=models.IntegerField(default=0),
        ),
        migrations.AddField(
            model_name='connectionapi',
            name='query_timeout',
            field=models.IntegerField(default=0),
        ),
    ]



from isapilib.api.models import BranchAPI
from isapilib.core.exceptions import PermissionDenied


def add_conn(user, branch: BranchAPI):
    name = branch.connection.create_connection()
    if not branch.check_permissions(user):
        raise PermissionDenied(f'You do not have permissions on this connection [{user.username}, {branch.pk}]')
    return name



from django.conf import settings

from isapilib.api.models import BranchAPI
from isapilib.auth.permissions import IsapilibPermission
from isapilib.core.exceptions import RequestError
from isapilib.core.utilities import is_test, get_dealer_field
from isapilib.external.connection import add_conn
from isapilib.logging import logger


class BaseRouter:
    default_database = (
        'ConnectionAPI'.lower(),
        'BranchAPI'.lower(),
        'UserAPI'.lower(),
        'PermissionAPI'.lower(),
        'ApiLogs'.lower(),
        'idtoken',
        'accesstoken',
        'refreshtoken',
        'application',
    )

    def get_branch(self, user, request) -> BranchAPI:
        if request.user.branch is None:
            raise RequestError(f"User don't have a default branch ({request.user})")
        return request.user.branch

    def external_db(self, request):
        if name := getattr(request, '_isapi_external_db', None):
            return name

        branch = self.get_branch(request.user, request)
        logger.debug('External database, request %s user %s, branch %s', id(request), request.user.pk, branch.pk)
        name = add_conn(request.user, branch)
        setattr(request, '_isapi_external_db', name)
        return name

    def _get_model_name(self, model):
        try:
            return model._meta.model_name
        except Exception:
            return 'execution'

    def db_for_read(self, model, **hints):
        model_name = self._get_model_name(model)
        if model_name in self.default_database:
            return 'default'

        request = IsapilibPermission.get_current_request()
        alias = self.external_db(request)

        logger.info('Database for read, model %s, alias %s, request %s', model_name, alias, id(request))
        return alias

    def db_for_write(self, model, **hints):
        model_name = self._get_model_name(model)
        if model_name in self.default_database:
            return 'default'

        request = IsapilibPermission.get_current_request()
        alias = self.external_db(request)

        logger.info("Database for write, model %s, alias %s, request %s", model_name, alias, id(request))
        return alias

    def allow_relation(self, *args, **kwargs):
        return True


def get_branch(dealer_id: str, src='header') -> BranchAPI:
    field = get_dealer_field()
    if dealer_id is None: raise RequestError(f'{field} {src} is required')
    branch = BranchAPI.objects.exclude(gwmbac='').filter(gwmbac=dealer_id, gwmbac__isnull=False).first()
    if branch is None: raise RequestError(f'El DealerID ({dealer_id}) no existe')
    return branch


class DealerRouter(BaseRouter):
    field = get_dealer_field()

    def get_branch(self, user, request) -> BranchAPI:
        if is_test():
            dealer_id = getattr(settings, self.field)
        else:
            dealer_id = request.headers.get(self.field)
        return get_branch(dealer_id)


class DealerBodyRouter(BaseRouter):
    field = get_dealer_field()

    def get_branch(self, user, request) -> BranchAPI:
        if is_test():
            dealer_id = getattr(settings, self.field)
        else:
            dealer_id = request.data.get(self.field)
        return get_branch(dealer_id, src='body')
